# do not forget to keep a three-number version!!!
import datetime
_current_year_ = datetime.datetime.now().year

__GridCal_VERSION__ = "4.1.0"

url = 'https://github.com/SanPen/GridCal'

about_msg = "GridCal v" + str(__GridCal_VERSION__) + '\n\n'

about_msg += "GridCal has been carefully crafted since 2015 to serve as a platform for research and consultancy.\n\n"

about_msg += 'This program comes with ABSOLUTELY NO WARRANTY. \n'
about_msg += 'This is free software, and you are welcome to redistribute it under certain conditions; '

about_msg += "GridCal is licensed under the GNU general public license V.3. "
about_msg += 'See the license file for more details. \n\n'
about_msg += "The source of GridCal can be found at:\n" + url + "\n\n"

about_msg += 'Copyright (C) 2015-' + str(_current_year_) + '\nSantiago Peñate Vera\n'
about_msg += 'Michel Lavoie (Transformer automation)\n'
about_msg += 'Bengt Lüers (Better testing)\n'
about_msg += 'Josep Fanals Batllori (HELM)\n'
about_msg += 'Manuel Navarro Catalán (Better documentation)\n'
about_msg += 'Paul Schultz (Grid Generator)\n'



# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.

from GridCal.Engine.basic_structures import MIPSolvers
from GridCal.ThirdParty.pulp import *


class Opf:

    def __init__(self, numerical_circuit, solver: MIPSolvers = MIPSolvers.CBC):
        """
        Optimal power flow template class
        :param numerical_circuit: NumericalCircuit instance
        """
        self.numerical_circuit = numerical_circuit

        self.theta = None
        self.Pg = None
        self.Pb = None
        self.Pl = None
        self.E = None
        self.s_from = None
        self.s_to = None
        self.overloads = None
        self.rating = None
        self.load_shedding = None
        self.nodal_restrictions = None

        self.solver = solver

        self.problem = self.formulate()

    def formulate(self):
        """
        Formulate the AC OPF time series in the non-sequential fashion (all to the solver at once)
        :return: PuLP Problem instance
        """

        # declare problem
        problem = LpProblem(name='DC_OPF_Time_Series')

        return problem

    def solve(self):
        """
        Call PuLP to solve the problem
        """
        # self.problem.writeLP('OPF.lp')
        if self.solver == MIPSolvers.CBC:
            params = PULP_CBC_CMD(fracGap=0.00001, threads=None, msg=1)

        elif self.solver == MIPSolvers.SCIP:
            params = SCIP_CMD(msg=1)

        elif self.solver == MIPSolvers.CPLEX:
            params = CPLEX_CMD(msg=1)

        elif self.solver == MIPSolvers.GUROBI:
            params = GUROBI_CMD(msg=1)

        elif self.solver == MIPSolvers.XPRESS:
            params = XPRESS(msg=1)

        else:
            raise Exception('Solver not supported! ' + str(self.solver))

        self.problem.solve(params)

        return LpStatus[self.problem.status]

    @staticmethod
    def extract(arr, make_abs=False):
        """
        Extract values fro the 1D array of LP variables
        :param arr: 1D array of LP variables
        :param make_abs: substitute the result by its abs value
        :return: 1D numpy array
        """
        val = np.zeros(arr.shape)
        for i in range(val.shape[0]):
            val[i] = arr[i].value()
        if make_abs:
            val = np.abs(val)

        return val

    def get_voltage(self):
        """
        return the complex voltages (time, device)
        :return: 2D array
        """
        angles = self.extract(self.theta)
        return np.ones_like(angles) * np.exp(-1j * angles)

    def get_overloads(self):
        """
        return the branch overloads (time, device)
        :return: 2D array
        """
        return self.extract(self.overloads)

    def get_loading(self):
        """
        return the branch loading (time, device)
        :return: 2D array
        """
        return self.extract(self.s_from, make_abs=True) / (self.rating + 1e-12)

    def get_branch_power(self):
        """
        return the branch loading (time, device)
        :return: 2D array
        """
        return self.extract(self.s_from, make_abs=True) * self.numerical_circuit.Sbase

    def get_battery_power(self):
        """
        return the battery dispatch (time, device)
        :return: 2D array
        """
        return self.extract(self.Pb) * self.numerical_circuit.Sbase

    def get_generator_power(self):
        """
        return the generator dispatch (time, device)
        :return: 2D array
        """
        return self.extract(self.Pg) * self.numerical_circuit.Sbase

    def get_load_shedding(self):
        """
        return the load shedding (time, device)
        :return: 2D array
        """
        return self.extract(self.load_shedding) * self.numerical_circuit.Sbase

    def get_load_power(self):
        """
        return the load shedding (time, device)
        :return: 2D array
        """
        return self.extract(self.Pl) * self.numerical_circuit.Sbase

    def get_shadow_prices(self):
        """
        Extract values fro the 2D array of LP variables
        :return: 2D numpy array
        """
        val = np.zeros(self.nodal_restrictions.shape)
        for i in range(val.shape[0]):
            if self.nodal_restrictions[i].pi is not None:
                val[i] = - self.nodal_restrictions[i].pi
        return val.transpose()

    def converged(self):
        return True


class OpfTimeSeries:

    def __init__(self, numerical_circuit, start_idx, end_idx, solver: MIPSolvers=MIPSolvers.CBC):
        """

        :param numerical_circuit:
        :param start_idx:
        :param end_idx:
        """
        self.numerical_circuit = numerical_circuit
        self.start_idx = start_idx
        self.end_idx = end_idx
        self.solver = solver

        self.theta = None
        self.Pg = None
        self.Pb = None
        self.Pl = None
        self.E = None
        self.s_from = None
        self.s_to = None
        self.overloads = None
        self.rating = None
        self.load_shedding = None
        self.nodal_restrictions = None

        self.problem = self.formulate()

    def formulate(self):
        """
        Formulate the AC OPF time series in the non-sequential fashion (all to the solver at once)
        :return: PuLP Problem instance
        """

        # declare problem
        problem = LpProblem(name='DC_OPF_Time_Series')

        return problem

    def solve(self, msg=False):
        """
        Call PuLP to solve the problem
        """

        if self.solver == MIPSolvers.CBC:
            params = PULP_CBC_CMD(fracGap=0.00001, threads=None, msg=msg)

        elif self.solver == MIPSolvers.SCIP:
            params = SCIP_CMD(msg=msg)

        elif self.solver == MIPSolvers.CPLEX:
            params = CPLEX_CMD(msg=msg)

        elif self.solver == MIPSolvers.GUROBI:
            params = GUROBI_CMD(msg=msg)

        elif self.solver == MIPSolvers.XPRESS:
            params = XPRESS(msg=msg)

        else:
            raise Exception('Solver not supported! ' + str(self.solver))

        self.problem.solve(params)

        return LpStatus[self.problem.status]

    @staticmethod
    def extract2D(arr, make_abs=False):
        """
        Extract values fro the 2D array of LP variables
        :param arr: 2D array of LP variables
        :param make_abs: substitute the result by its abs value
        :return: 2D numpy array
        """
        val = np.zeros(arr.shape)
        for i, j in product(range(val.shape[0]), range(val.shape[1])):
            val[i, j] = arr[i, j].value()
        if make_abs:
            val = np.abs(val)

        return val

    def get_voltage(self):
        """
        return the complex voltages (time, device)
        :return: 2D array
        """
        angles = self.extract2D(self.theta)
        return np.ones_like(angles) * np.exp(-1j * angles)

    def get_overloads(self):
        """
        return the branch overloads (time, device)
        :return: 2D array
        """
        return self.extract2D(self.overloads)

    def get_loading(self):
        """
        return the branch loading (time, device)
        :return: 2D array
        """
        return self.extract2D(self.s_from, make_abs=True) / self.rating

    def get_branch_power(self):
        """
        return the branch loading (time, device)
        :return: 2D array
        """
        return self.extract2D(self.s_from, make_abs=True) * self.numerical_circuit.Sbase

    def get_battery_power(self):
        """
        return the battery dispatch (time, device)
        :return: 2D array
        """
        return self.extract2D(self.Pb) * self.numerical_circuit.Sbase

    def get_battery_energy(self):
        """
        return the battery energy (time, device)
        :return: 2D array
        """
        return self.extract2D(self.E) * self.numerical_circuit.Sbase

    def get_generator_power(self):
        """
        return the generator dispatch (time, device)
        :return: 2D array
        """
        return self.extract2D(self.Pg) * self.numerical_circuit.Sbase

    def get_load_shedding(self):
        """
        return the load shedding (time, device)
        :return: 2D array
        """
        return self.extract2D(self.load_shedding) * self.numerical_circuit.Sbase

    def get_load_power(self):
        """
        return the load shedding (time, device)
        :return: 2D array
        """
        return self.extract2D(self.Pl) * self.numerical_circuit.Sbase

    def get_shadow_prices(self):
        """
        Extract values fro the 2D array of LP variables
        :return: 2D numpy array
        """
        val = np.zeros(self.nodal_restrictions.shape)
        for i, j in product(range(val.shape[0]), range(val.shape[1])):
            if self.nodal_restrictions[i, j].pi is not None:
                val[i, j] = - self.nodal_restrictions[i, j].pi
        return val.transpose()



"""
Script to update correctly the main GUI (.py) file from the Qt design (.ui) file
"""
from subprocess import call

if __name__ == '__main__':
    # pyrcc5 icons.qrc -o icons_rc.py
    # pyuic5 -x MainWindow.ui -o MainWindow.py

    filename = 'gui.py'
    filename_ui = 'gui.ui'

    # update icon/images resources
    call(['pyside2-rcc', 'icons.qrc', '-o', 'icons_rc.py'])

    # update ui handler file
    call(['pyside2-uic', '-x', filename_ui, '-o', filename])


    # replace annoying text import
    # Read in the file
    with open(filename, 'r') as file:
        file_data = file.read()

    # Replace the target string
    file_data = file_data.replace('import icons_rc', 'from .icons_rc import *')

    # Write the file out again
    with open(filename, 'w') as file:
        file.write(file_data)



# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum
from datetime import datetime
from PySide2 import QtCore, QtGui, QtWidgets

from GridCal.Engine.basic_structures import Logger
from GridCal.Gui.GuiFunctions import ObjectsModel, get_tree_model


def get_list_model(lst, checks=False):
    """
    Pass a list to a list model
    """
    list_model = QtGui.QStandardItemModel()
    if lst is not None:
        if not checks:
            for val in lst:
                # for the list model
                item = QtGui.QStandardItem(str(val))
                item.setEditable(False)
                list_model.appendRow(item)
        else:
            for val in lst:
                # for the list model
                item = QtGui.QStandardItem(str(val))
                item.setEditable(False)
                item.setCheckable(True)
                item.setCheckState(QtCore.Qt.Checked)
                list_model.appendRow(item)

    return list_model


class ProfileTypes(Enum):
    Loads = 1,
    Generators = 2


class NewProfilesStructureDialogue(QtWidgets.QDialog):
    """
    New profile dialogue window
    """
    def __init__(self):
        super(NewProfilesStructureDialogue, self).__init__()
        self.setObjectName("self")
        # self.resize(200, 71)
        # self.setMinimumSize(QtCore.QSize(200, 71))
        # self.setMaximumSize(QtCore.QSize(200, 71))
        self.setContextMenuPolicy(QtCore.Qt.NoContextMenu)
        # icon = QtGui.QIcon()
        # icon.addPixmap(QtGui.QPixmap("Icons/Plus-32.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        # self.setWindowIcon(icon)
        self.layout = QtWidgets.QVBoxLayout(self)

        # calendar
        self.calendar = QtWidgets.QDateTimeEdit()
        d = datetime.today()
        self.calendar.setDateTime(QtCore.QDateTime(d.year, 1, 1, 00, 00, 00))

        # number of time steps
        self.steps_spinner = QtWidgets.QSpinBox()
        self.steps_spinner.setMinimum(1)
        self.steps_spinner.setMaximum(9999999)
        self.steps_spinner.setValue(1)

        # time step length
        self.step_length = QtWidgets.QDoubleSpinBox()
        self.step_length.setMinimum(1)
        self.step_length.setMaximum(60)
        self.step_length.setValue(1)

        # units combo box
        self.units = QtWidgets.QComboBox()
        self.units.setModel(get_list_model(['h', 'm', 's']))

        # accept button
        self.accept_btn = QtWidgets.QPushButton()
        self.accept_btn.setText('Accept')
        self.accept_btn.clicked.connect(self.accept_click)

        # labels

        # add all to the GUI
        self.layout.addWidget(QtWidgets.QLabel("Start date"))
        self.layout.addWidget(self.calendar)

        self.layout.addWidget(QtWidgets.QLabel("Number of time steps"))
        self.layout.addWidget(self.steps_spinner)

        self.layout.addWidget(QtWidgets.QLabel("Time step length"))
        self.layout.addWidget(self.step_length)

        self.layout.addWidget(QtWidgets.QLabel("Time units"))
        self.layout.addWidget(self.units)

        self.layout.addWidget(self.accept_btn)

        self.setLayout(self.layout)

        self.setWindowTitle('New profiles structure')

    def accept_click(self):
        self.accept()

    def get_values(self):
        steps = self.steps_spinner.value()

        step_length = self.step_length.value()

        step_unit = self.units.currentText()

        time_base = self.calendar.dateTime()

        return steps, step_length, step_unit, time_base.toPython()

#
# class LogsModel(QtCore.QAbstractTableModel):
#
#     def __init__(self, logs: Logger, parent=None):
#
#         QtCore.QAbstractTableModel.__init__(self, parent)
#
#         self.logs = logs
#
#     def flags(self, index):
#         return QtCore.Qt.ItemIsEnabled
#
#     def rowCount(self, parent=None):
#         return len(self.logs)
#
#     def columnCount(self, parent=None):
#         return 2
#
#     def headerData(self, section, orientation, role=None):
#         """
#
#         :param section:
#         :param orientation:
#         :param role:
#         :return:
#         """
#         if role == QtCore.Qt.DisplayRole and orientation == QtCore.Qt.Horizontal:
#             if section == 0:
#                 return "Severity"
#             elif section == 1:
#                 return "Message"
#         return None
#
#     def data(self, index, role=None):
#         """
#
#         :param index:
#         :param role:
#         :return:
#         """
#         if index.isValid() and role == QtCore.Qt.DisplayRole:
#
#             if index.column() == 0:
#                 return str(self.logs.severity[index.row()])
#
#             elif index.column() == 1:
#                 return str(self.logs.messages[index.row()])
#
#         return None


def fill_tree_from_logs(logger: Logger):
    """
    Fill logger tree
    :param logger: Logger instance
    :return: QStandardItemModel instance
    """
    d = logger.to_dict()
    editable = False
    model = QtGui.QStandardItemModel()
    model.setHorizontalHeaderLabels(['Time', 'Element', 'Value', 'Expected value'])
    parent = model.invisibleRootItem()

    for severity, messages_dict in d.items():
        severity_child = QtGui.QStandardItem(severity)

        # print(severity)

        for message, data_list in messages_dict.items():
            message_child = QtGui.QStandardItem(message)

            # print('\t', message)

            for time, elm, value, expected_value in data_list:

                # print('\t', '\t', time, elm, value, expected_value)

                time_child = QtGui.QStandardItem(time)
                time_child.setEditable(editable)

                elm_child = QtGui.QStandardItem(elm)
                elm_child.setEditable(editable)

                value_child = QtGui.QStandardItem(value)
                value_child.setEditable(editable)

                expected_val_child = QtGui.QStandardItem(expected_value)
                expected_val_child.setEditable(editable)

                message_child.appendRow([time_child, elm_child, value_child, expected_val_child])

            message_child.setEditable(editable)
            severity_child.appendRow(message_child)

        severity_child.setEditable(editable)
        parent.appendRow(severity_child)

    return model


class LogsDialogue(QtWidgets.QDialog):
    """
    New profile dialogue window
    """
    def __init__(self, name, logs: Logger()):
        super(LogsDialogue, self).__init__()
        self.setObjectName("self")
        self.setContextMenuPolicy(QtCore.Qt.NoContextMenu)
        self.main_layout = QtWidgets.QVBoxLayout(self)

        self.logs = logs

        # logs_list
        self.logs_table = QtWidgets.QTreeView()
        model = fill_tree_from_logs(logs)
        self.logs_table.setModel(model)
        self.logs_table.setFirstColumnSpanned(0, QtCore.QModelIndex(), True)
        self.logs_table.setFirstColumnSpanned(1, QtCore.QModelIndex(), True)
        self.logs_table.setAnimated(True)

        # accept button
        self.accept_btn = QtWidgets.QPushButton()
        self.accept_btn.setText('Accept')
        self.accept_btn.clicked.connect(self.accept_click)

        self.save_btn = QtWidgets.QPushButton()
        self.save_btn.setText('Save')
        self.save_btn.clicked.connect(self.save_click)

        self.btn_frame = QtWidgets.QFrame()
        self.btn_layout = QtWidgets.QHBoxLayout(self.btn_frame)
        self.btn_spacer = QtWidgets.QSpacerItem(10, 10, QtWidgets.QSizePolicy.Expanding)
        self.btn_layout.addWidget(self.save_btn)
        self.btn_layout.addSpacerItem(self.btn_spacer)
        self.btn_layout.addWidget(self.accept_btn)
        self.btn_frame.setLayout(self.btn_layout)

        # add all to the GUI
        self.main_layout.addWidget(self.logs_table)
        self.main_layout.addWidget(self.btn_frame)

        self.setLayout(self.main_layout)

        self.setWindowTitle(name)

        h = 400
        self.resize(int(1.61 * h), h)

    def accept_click(self):
        """
        Accept and close
        """
        self.accept()

    def save_click(self):
        """
        Save the logs to excel or CSV
        """
        file, filter = QtWidgets.QFileDialog.getSaveFileName(self, "Export results", '',
                                                             filter="CSV (*.csv);;Excel files (*.xlsx)",)

        if file != '':
            if 'xlsx' in filter:
                f = file
                if not f.endswith('.xlsx'):
                    f += '.xlsx'
                self.logs.to_xlsx(f)

            if 'csv' in filter:
                f = file
                if not f.endswith('.csv'):
                    f += '.csv'
                self.logs.to_csv(f)


class ElementsDialogue(QtWidgets.QDialog):
    """
    Selected elements dialogue window
    """

    def __init__(self, name, elements: list()):
        super(ElementsDialogue, self).__init__()
        self.setObjectName("self")
        self.setContextMenuPolicy(QtCore.Qt.NoContextMenu)
        self.layout = QtWidgets.QVBoxLayout(self)

        # build elements list
        self.objects_table = QtWidgets.QTableView()

        if len(elements) > 0:
            model = ObjectsModel(elements, elements[0].editable_headers,
                                 parent=self.objects_table, editable=False,
                                 non_editable_attributes=[1, 2, 14])

            self.objects_table.setModel(model)

        # accept button
        self.accept_btn = QtWidgets.QPushButton()
        self.accept_btn.setText('Proceed')
        self.accept_btn.clicked.connect(self.accept_click)

        # Copy button
        self.copy_btn = QtWidgets.QPushButton()
        self.copy_btn.setText('Copy')
        self.copy_btn.clicked.connect(self.copy_click)

        # add all to the GUI
        self.layout.addWidget(self.objects_table)
        self.frame2 = QtWidgets.QFrame()
        self.layout.addWidget(self.frame2)
        self.layout2 = QtWidgets.QHBoxLayout(self.frame2)

        self.layout2.addWidget(self.accept_btn)
        # self.layout2.addWidget(QtWidgets.QSpacerItem())
        self.layout2.addWidget(self.copy_btn)

        self.setLayout(self.layout)

        self.setWindowTitle(name)

        self.accepted = False

    def accept_click(self):
        self.accepted = True
        self.accept()

    def copy_click(self):
        pass


if __name__ == "__main__":
    import sys
    from PySide2.QtWidgets import QApplication
    app = QApplication(sys.argv)
    window = LogsDialogue(name='', logs=Logger())
    window.show()
    sys.exit(app.exec_())



# Copyright (c) 1996-2015 PSERC. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE_MATPOWER file.

# Copyright 1996-2015 PSERC. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

# Copyright (c) 2016-2017 by University of Kassel and Fraunhofer Institute for Wind Energy and
# Energy System Technology (IWES), Kassel. All rights reserved. Use of this source code is governed
# by a BSD-style license that can be found in the LICENSE file.

# The file has been modified from Pypower.
# The function mu() has been added to the solver in order to provide an optimal iteration control
#
# Copyright (c) 2018 Santiago Peñate Vera
#
# This file retains the BSD-Style license


from numpy import array, angle, exp, linalg, r_, Inf, conj, diag, asmatrix, asarray, zeros_like, zeros, complex128, \
empty, float64, int32, arange
from scipy.sparse import csr_matrix as sparse, hstack, vstack
from scipy.sparse.linalg import spsolve, splu
import numpy as np
import pandas as pd
import numba as nb
import time
from warnings import warn
from scipy.sparse import coo_matrix, csc_matrix
from scipy.sparse import hstack as hs, vstack as vs
from scipy.sparse.linalg import factorized, spsolve
from matplotlib import pyplot as plt
import scipy
scipy.ALLOW_THREADS = True
import time
import numpy as np

np.set_printoptions(precision=8, suppress=True, linewidth=320)


def dSbus_dV(Ybus, V, I):
    """
    Computes partial derivatives of power injection w.r.t. voltage.
    :param Ybus: Admittance matrix
    :param V: Bus voltages array
    :param I: Bus current injections array
    :return:
    """
    '''
    Computes partial derivatives of power injection w.r.t. voltage.

    Returns two matrices containing partial derivatives of the complex bus
    power injections w.r.t voltage magnitude and voltage angle respectively
    (for all buses). If C{Ybus} is a sparse matrix, the return values will be
    also. The following explains the expressions used to form the matrices::

        Ibus = Ybus * V - I

        S = diag(V) * conj(Ibus) = diag(conj(Ibus)) * V

    Partials of V & Ibus w.r.t. voltage magnitudes::
        dV/dVm = diag(V / abs(V))
        dI/dVm = Ybus * dV/dVm = Ybus * diag(V / abs(V))

    Partials of V & Ibus w.r.t. voltage angles::
        dV/dVa = j * diag(V)
        dI/dVa = Ybus * dV/dVa = Ybus * j * diag(V)

    Partials of S w.r.t. voltage magnitudes::
        dS/dVm = diag(V) * conj(dI/dVm) + diag(conj(Ibus)) * dV/dVm
               = diag(V) * conj(Ybus * diag(V / abs(V)))
                                        + conj(diag(Ibus)) * diag(V / abs(V))

    Partials of S w.r.t. voltage angles::
        dS/dVa = diag(V) * conj(dI/dVa) + diag(conj(Ibus)) * dV/dVa
               = diag(V) * conj(Ybus * j * diag(V))
                                        + conj(diag(Ibus)) * j * diag(V)
               = -j * diag(V) * conj(Ybus * diag(V))
                                        + conj(diag(Ibus)) * j * diag(V)
               = j * diag(V) * conj(diag(Ibus) - Ybus * diag(V))

    For more details on the derivations behind the derivative code used
    in PYPOWER information, see:

    [TN2]  R. D. Zimmerman, "AC Power Flows, Generalized OPF Costs and
    their Derivatives using Complex Matrix Notation", MATPOWER
    Technical Note 2, February 2010.
    U{http://www.pserc.cornell.edu/matpower/TN2-OPF-Derivatives.pdf}

    @author: Ray Zimmerman (PSERC Cornell)
    '''

    ib = range(len(V))

    Ibus = Ybus * V - I

    diagV = sparse((V, (ib, ib)))
    diagIbus = sparse((Ibus, (ib, ib)))
    diagVnorm = sparse((V / np.abs(V), (ib, ib)))

    dS_dVm = diagV * conj(Ybus * diagVnorm) + conj(diagIbus) * diagVnorm
    dS_dVa = 1.0j * diagV * conj(diagIbus - Ybus * diagV)

    return dS_dVm, dS_dVa


def mu(Ybus, Ibus, J, incS, dV, dx, pvpq, pq):
    """
    Calculate the Iwamoto acceleration parameter as described in:
    "A Load Flow Calculation Method for Ill-Conditioned Power Systems" by Iwamoto, S. and Tamura, Y."
    Args:
        Ybus: Admittance matrix
        J: Jacobian matrix
        incS: mismatch vector
        dV: voltage increment (in complex form)
        dx: solution vector as calculated dx = solve(J, incS)
        pvpq: array of the pq and pv indices
        pq: array of the pq indices

    Returns:
        the Iwamoto's optimal multiplier for ill conditioned systems
    """
    # evaluate the Jacobian of the voltage derivative
    # theoretically this is the second derivative matrix
    # since the Jacobian (J2) has been calculated with dV instead of V
    J2 = Jacobian(Ybus, dV, Ibus, pq, pvpq)

    a = incS
    b = J * dx
    c = 0.5 * dx * J2 * dx

    g0 = -a.dot(b)
    g1 = b.dot(b) + 2 * a.dot(c)
    g2 = -3.0 * b.dot(c)
    g3 = 2.0 * c.dot(c)

    roots = np.roots([g3, g2, g1, g0])
    # three solutions are provided, the first two are complex, only the real solution is valid
    return roots[2].real


def Jacobian(Ybus, V, Ibus, pq, pvpq):
    """
    Computes the system Jacobian matrix
    Args:
        Ybus: Admittance matrix
        V: Array of nodal voltages
        Ibus: Array of nodal current injections
        pq: Array with the indices of the PQ buses
        pvpq: Array with the indices of the PV and PQ buses

    Returns:
        The system Jacobian matrix
    """
    dS_dVm, dS_dVa = dSbus_dV(Ybus, V, Ibus)  # compute the derivatives

    J11 = dS_dVa[array([pvpq]).T, pvpq].real
    J12 = dS_dVm[array([pvpq]).T, pq].real
    J21 = dS_dVa[array([pq]).T, pvpq].imag
    J22 = dS_dVm[array([pq]).T, pq].imag

    J = vstack([hstack([J11, J12]),
                hstack([J21, J22])], format="csr")

    return J


def NR(Ybus, Sbus, V0, Ibus, pv, pq, tol, max_it=15, mu0=0.05, error_registry=None):
    """
    Solves the power flow using a full Newton's method
    Args:
        Ybus: Admittance matrix
        Sbus: Array of nodal power injections
        V0: Array of nodal voltages (initial solution)
        Ibus: Array of nodal current injections
        pv: Array with the indices of the PV buses
        pq: Array with the indices of the PQ buses
        tol: Tolerance
        max_it: Maximum number of iterations
        mu0: parameter used to correct the "bad" iterations, should be be between 1e-3 ~ 0.5
        error_registry: list to store the error for plotting
    Returns:
        Voltage solution, converged?, error, calculated power injections

    @Author: Santiago Penate Vera
    """
    start = time.time()

    # initialize
    back_track_counter = 0
    back_track_iterations = 0
    converged = 0
    iter_ = 0
    V = V0
    Va = np.angle(V)
    Vm = np.abs(V)
    dVa = np.zeros_like(Va)
    dVm = np.zeros_like(Vm)

    # set up indexing for updating V
    pvpq = r_[pv, pq]
    npv = len(pv)
    npq = len(pq)

    # j1:j2 - V angle of pv buses
    j1 = 0
    j2 = npv
    # j3:j4 - V angle of pq buses
    j3 = j2
    j4 = j2 + npq
    # j5:j6 - V mag of pq buses
    j5 = j4
    j6 = j4 + npq

    # evaluate F(x0)
    Scalc = V * conj(Ybus * V - Ibus)
    dS = Scalc - Sbus  # compute the mismatch
    f = r_[dS[pv].real, dS[pq].real, dS[pq].imag]

    # check tolerance
    norm_f = 0.5 * f.dot(f)

    if error_registry is not None:
        error_registry.append(norm_f)

    if norm_f < tol:
        converged = 1

    # do Newton iterations
    while not converged and iter_ < max_it:
        # update iteration counter
        iter_ += 1

        # evaluate Jacobian
        J = Jacobian(Ybus, V, Ibus, pq, pvpq)

        # compute update step
        dx = spsolve(J, f)

        # reassign the solution vector
        if npv:
            dVa[pv] = dx[j1:j2]
        if npq:
            dVa[pq] = dx[j3:j4]
            dVm[pq] = dx[j5:j6]

        # update voltage the Newton way (mu=1)
        mu_ = mu0
        Vm -= mu_ * dVm
        Va -= mu_ * dVa
        V = Vm * exp(1j * Va)

        Vm = np.abs(V)  # update Vm and Va again in case
        Va = np.angle(V)  # we wrapped around with a negative Vm

        # compute the mismatch function f(x_new)
        dS = V * conj(Ybus * V - Ibus) - Sbus  # complex power mismatch
        f_new = r_[dS[pv].real, dS[pq].real, dS[pq].imag]  # concatenate to form the mismatch function
        norm_f = 0.5 * f_new.dot(f_new)

        if error_registry is not None:
            error_registry.append(norm_f)

        if norm_f < tol:
            converged = 1

    end = time.time()
    elapsed = end - start

    print('iter_', iter_, '  -  back_track_counter', back_track_counter,
          '  -  back_track_iterations', back_track_iterations)

    return V, converged, norm_f, Scalc, elapsed


def NR_LS1(Ybus, Sbus, V0, Ibus, pv, pq, tol, max_it=15, acceleration_parameter=0.05, error_registry=None):
    """
    Solves the power flow using a full Newton's method with the Iwamoto optimal step factor.
    Args:
        Ybus: Admittance matrix
        Sbus: Array of nodal power injections
        V0: Array of nodal voltages (initial solution)
        Ibus: Array of nodal current injections
        pv: Array with the indices of the PV buses
        pq: Array with the indices of the PQ buses
        tol: Tolerance
        max_it: Maximum number of iterations
        acceleration_parameter: parameter used to correct the "bad" iterations, should be be between 1e-3 ~ 0.5
        error_registry: list to store the error for plotting
    Returns:
        Voltage solution, converged?, error, calculated power injections

    @Author: Santiago Penate Vera
    """
    start = time.time()

    # initialize
    back_track_counter = 0
    back_track_iterations = 0
    alpha = 1e-4
    converged = 0
    iter_ = 0
    V = V0
    Va = angle(V)
    Vm = abs(V)
    dVa = zeros_like(Va)
    dVm = zeros_like(Vm)

    # set up indexing for updating V
    pvpq = r_[pv, pq]
    npv = len(pv)
    npq = len(pq)

    # j1:j2 - V angle of pv buses
    j1 = 0
    j2 = npv
    # j3:j4 - V angle of pq buses
    j3 = j2
    j4 = j2 + npq
    # j5:j6 - V mag of pq buses
    j5 = j4
    j6 = j4 + npq

    # evaluate F(x0)
    Scalc = V * conj(Ybus * V - Ibus)
    dS = Scalc - Sbus  # compute the mismatch
    f = r_[dS[pv].real, dS[pq].real, dS[pq].imag]

    # check tolerance
    norm_f = linalg.norm(f, Inf)

    if error_registry is not None:
        error_registry.append(norm_f)

    if norm_f < tol:
        converged = 1

    # do Newton iterations
    while not converged and iter_ < max_it:
        # update iteration counter
        iter_ += 1

        # evaluate Jacobian
        J = Jacobian(Ybus, V, Ibus, pq, pvpq)

        # compute update step
        dx = spsolve(J, f)

        # reassign the solution vector
        if npv:
            dVa[pv] = dx[j1:j2]
        if npq:
            dVa[pq] = dx[j3:j4]
            dVm[pq] = dx[j5:j6]

        # update voltage the Newton way (mu=1)
        mu_ = 1.0
        Vm -= mu_ * dVm
        Va -= mu_ * dVa
        Vnew = Vm * exp(1j * Va)

        # compute the mismatch function f(x_new)
        dS = Vnew * conj(Ybus * Vnew - Ibus) - Sbus  # complex power mismatch
        f_new = r_[dS[pv].real, dS[pq].real, dS[pq].imag]  # concatenate to form the mismatch function
        norm_f_prev = linalg.norm(f + alpha * (f * J).dot(f_new - f), Inf)

        if error_registry is not None:
            error_registry.append(norm_f_prev)

        cond = norm_f < norm_f_prev  # condition to back track (no improvement at all)

        if not cond:
            back_track_counter += 1

        l_iter = 0
        while not cond and l_iter < 10 and mu_ > 0.01:
            # line search back

            # to divide mu by 4 is the simplest backtracking process
            # TODO: implement the more complex mu backtrack from numerical recipes

            # update voltage with a closer value to the last value in the Jacobian direction
            mu_ *= acceleration_parameter
            Vm -= mu_ * dVm
            Va -= mu_ * dVa
            Vnew = Vm * exp(1j * Va)

            # compute the mismatch function f(x_new)
            dS = Vnew * conj(Ybus * Vnew - Ibus) - Sbus  # complex power mismatch
            f_new = r_[dS[pv].real, dS[pq].real, dS[pq].imag]  # concatenate to form the mismatch function

            norm_f_new = linalg.norm(f_new, Inf)
            norm_f_new_prev = linalg.norm(f + alpha * (f * J).dot(f_new - f), Inf)

            cond = norm_f_new < norm_f_new_prev

            if error_registry is not None:
                error_registry.append(norm_f_new_prev)

            l_iter += 1
            back_track_iterations += 1

        # update calculation variables
        V = Vnew
        f = f_new

        # check for convergence
        norm_f = linalg.norm(f, Inf)

        if error_registry is not None:
            error_registry.append(norm_f)

        if norm_f < tol:
            converged = 1

    end = time.time()
    elapsed = end - start

    print('iter_', iter_, '  -  back_track_counter', back_track_counter,
          '  -  back_track_iterations', back_track_iterations)

    return V, converged, norm_f, Scalc, elapsed


def NR_LS2(Ybus, Sbus, V0, Ibus, pv, pq, tol, max_it=15, acceleration_parameter=0.05, error_registry=None):
    """
    Solves the power flow using a full Newton's method with the Iwamoto optimal step factor.
    Args:
        Ybus: Admittance matrix
        Sbus: Array of nodal power injections
        V0: Array of nodal voltages (initial solution)
        Ibus: Array of nodal current injections
        pv: Array with the indices of the PV buses
        pq: Array with the indices of the PQ buses
        tol: Tolerance
        max_it: Maximum number of iterations
        acceleration_parameter: parameter used to correct the "bad" iterations, should be be between 1e-3 ~ 0.5
        error_registry: list to store the error for plotting
    Returns:
        Voltage solution, converged?, error, calculated power injections

    @Author: Santiago Penate Vera
    """
    start = time.time()

    # initialize
    back_track_counter = 0
    back_track_iterations = 0
    converged = 0
    iter_ = 0
    V = V0
    Va = np.angle(V)
    Vm = np.abs(V)
    dVa = np.zeros_like(Va)
    dVm = np.zeros_like(Vm)

    # set up indexing for updating V
    pvpq = r_[pv, pq]
    npv = len(pv)
    npq = len(pq)

    # j1:j2 - V angle of pv buses
    j1 = 0
    j2 = npv
    # j3:j4 - V angle of pq buses
    j3 = j2
    j4 = j2 + npq
    # j5:j6 - V mag of pq buses
    j5 = j4
    j6 = j4 + npq

    # evaluate F(x0)
    Scalc = V * conj(Ybus * V - Ibus)
    dS = Scalc - Sbus  # compute the mismatch
    f = r_[dS[pv].real, dS[pq].real, dS[pq].imag]

    # check tolerance
    norm_f = 0.5 * f.dot(f)

    if error_registry is not None:
        error_registry.append(norm_f)

    if norm_f < tol:
        converged = 1

    # do Newton iterations
    while not converged and iter_ < max_it:
        # update iteration counter
        iter_ += 1

        # evaluate Jacobian
        J = Jacobian(Ybus, V, Ibus, pq, pvpq)

        # compute update step
        dx = spsolve(J, f)

        # reassign the solution vector
        if npv:
            dVa[pv] = dx[j1:j2]
        if npq:
            dVa[pq] = dx[j3:j4]
            dVm[pq] = dx[j5:j6]

        # update voltage the Newton way (mu=1)
        mu_ = 1.0
        Vm -= mu_ * dVm
        Va -= mu_ * dVa
        Vnew = Vm * exp(1j * Va)

        # compute the mismatch function f(x_new)
        dS = Vnew * conj(Ybus * Vnew - Ibus) - Sbus  # complex power mismatch
        f_new = r_[dS[pv].real, dS[pq].real, dS[pq].imag]  # concatenate to form the mismatch function
        norm_f_new = 0.5 * f_new.dot(f_new)

        if error_registry is not None:
            error_registry.append(norm_f_new)

        cond = norm_f_new > norm_f  # condition to back track (no improvement at all)

        if not cond:
            back_track_counter += 1

        l_iter = 0
        while not cond and l_iter < 10 and mu_ > 0.01:
            # line search back
            # update voltage with a closer value to the last value in the Jacobian direction
            mu_ *= acceleration_parameter
            Vm -= mu_ * dVm
            Va -= mu_ * dVa
            Vnew = Vm * exp(1.0j * Va)

            # compute the mismatch function f(x_new)
            dS = Vnew * conj(Ybus * Vnew - Ibus) - Sbus  # complex power mismatch
            f_new = r_[dS[pv].real, dS[pq].real, dS[pq].imag]  # concatenate to form the mismatch function

            norm_f_new = 0.5 * f_new.dot(f_new)

            cond = norm_f_new > norm_f

            if error_registry is not None:
                error_registry.append(norm_f_new)

            l_iter += 1
            back_track_iterations += 1

        # update calculation variables
        V = Vnew
        f = f_new

        # check for convergence
        norm_f = 0.5 * f_new.dot(f_new)

        if error_registry is not None:
            error_registry.append(norm_f)

        if norm_f < tol:
            converged = 1

    end = time.time()
    elapsed = end - start

    print('iter_', iter_, '  -  back_track_counter', back_track_counter,
          '  -  back_track_iterations', back_track_iterations)

    return V, converged, norm_f, Scalc, elapsed


def NR_LS3(Ybus, Sbus, V0, Ibus, pv, pq, vd, tol, bus_installed_power,
           max_it=15, acceleration_parameter=0.05, error_registry=None, correct_slack=True):
    """
    Solves the power flow using a full Newton's method with the Iwamoto optimal step factor.
    Args:
        Ybus: Admittance matrix
        Sbus: Array of nodal power injections
        V0: Array of nodal voltages (initial solution)
        Ibus: Array of nodal current injections
        pv: Array with the indices of the PV buses
        pq: Array with the indices of the PQ buses
        tol: Tolerance
        max_it: Maximum number of iterations
        acceleration_parameter: parameter used to correct the "bad" iterations, should be be between 1e-3 ~ 0.5
        error_registry: list to store the error for plotting
    Returns:
        Voltage solution, converged?, error, calculated power injections

    @Author: Santiago Penate Vera
    """
    start = time.time()

    # initialize
    back_track_counter = 0
    back_track_iterations = 0
    alpha = 1e-4
    converged = 0
    iter_ = 0
    V = V0
    Va = angle(V)
    Vm = abs(V)
    dVa = zeros_like(Va)
    dVm = zeros_like(Vm)
    slack_corrected = False

    # set up indexing for updating V
    pvpq = r_[pv, pq]
    npv = len(pv)
    npq = len(pq)

    # j1:j2 - V angle of pv buses
    j1 = 0
    j2 = npv
    # j3:j4 - V angle of pq buses
    j3 = j2
    j4 = j2 + npq
    # j5:j6 - V mag of pq buses
    j5 = j4
    j6 = j4 + npq

    # evaluate F(x0)
    Scalc = V * conj(Ybus * V - Ibus)
    dS = Scalc - Sbus  # compute the mismatch
    f = r_[dS[pv].real, dS[pq].real, dS[pq].imag]

    # check tolerance
    norm_f = linalg.norm(f, Inf)

    if error_registry is not None:
        error_registry.append(norm_f)

    if norm_f < tol:
        converged = 1

    # do Newton iterations
    while not converged and iter_ < max_it:
        # update iteration counter
        iter_ += 1

        # evaluate Jacobian
        J = Jacobian(Ybus, V, Ibus, pq, pvpq)

        # compute update step
        dx = spsolve(J, f)

        # reassign the solution vector
        if npv:
            dVa[pv] = dx[j1:j2]
        if npq:
            dVa[pq] = dx[j3:j4]
            dVm[pq] = dx[j5:j6]

        # update voltage the Newton way (mu=1)
        mu_ = 1.0
        Vm -= mu_ * dVm
        Va -= mu_ * dVa
        Vnew = Vm * exp(1j * Va)

        # compute the mismatch function f(x_new)
        dS = Vnew * conj(Ybus * Vnew - Ibus) - Sbus  # complex power mismatch
        f_new = r_[dS[pv].real, dS[pq].real, dS[pq].imag]  # concatenate to form the mismatch function
        norm_f_prev = linalg.norm(f + alpha * (f * J).dot(f_new - f), Inf)

        if error_registry is not None:
            error_registry.append(norm_f_prev)

        cond = norm_f < norm_f_prev  # condition to back track (no improvement at all)

        if not cond:
            back_track_counter += 1

        l_iter = 0
        while not cond and l_iter < 10 and mu_ > 0.01:
            # line search back

            # to divide mu by 4 is the simplest backtracking process
            # TODO: implement the more complex mu backtrack from numerical recipes

            # update voltage with a closer value to the last value in the Jacobian direction
            mu_ *= acceleration_parameter
            Vm -= mu_ * dVm
            Va -= mu_ * dVa
            Vnew = Vm * exp(1j * Va)

            # compute the mismatch function f(x_new)
            dS = Vnew * conj(Ybus * Vnew - Ibus) - Sbus  # complex power mismatch
            f_new = r_[dS[pv].real, dS[pq].real, dS[pq].imag]  # concatenate to form the mismatch function

            norm_f_new = linalg.norm(f_new, Inf)
            norm_f_new_prev = linalg.norm(f + alpha * (f * J).dot(f_new - f), Inf)

            cond = norm_f_new < norm_f_new_prev

            if error_registry is not None:
                error_registry.append(norm_f_new_prev)

            l_iter += 1
            back_track_iterations += 1

        # update calculation variables
        V = Vnew
        f = f_new

        # correct slack
        if correct_slack:
            if norm_f <= 0.01 and not slack_corrected:
                Svd = V[vd] * np.conj(Ybus[vd, :].dot(V))
                slack_power = Svd.real.sum()
                total_installed_power = bus_installed_power.sum()
                if total_installed_power > 0.0:
                    delta = slack_power * bus_installed_power / total_installed_power
                    Sbus += delta
                    slack_corrected = True
                    print('Corrected slack!', delta.sum() * 100, 'MW')

        # check for convergence
        norm_f = linalg.norm(f, Inf)

        if error_registry is not None:
            error_registry.append(norm_f)

        if norm_f < tol:
            converged = 1

    end = time.time()
    elapsed = end - start

    print('iter_', iter_, '  -  back_track_counter', back_track_counter,
          '  -  back_track_iterations', back_track_iterations)

    if slack_corrected:
        Svd = V[vd] * np.conj(Ybus[vd, :].dot(V))
        slack_power = Svd.real.sum()
        print('Final slack', slack_power.real * 100, 'MW')

    return V, converged, norm_f, Scalc, elapsed


def F(V, Ybus, S, I, pq, pv):
    """

    :param V:
    :param Ybus:
    :param S:
    :param I:
    :param pq:
    :param pv:
    :return:
    """

    # compute the mismatch function f(x_new)
    dS = V * np.conj(Ybus * V - I) - S

    return np.r_[dS[pv].real, dS[pq].real, dS[pq].imag]  # concatenate to form the mismatch function


def fx(x, Ybus, S, I, pq, pv, pvpq, j1, j2, j3, j4, j5, j6, Va, Vm):
    """

    :param x:
    :param Ybus:
    :param S:
    :param I:
    :param pq:
    :param pv:
    :param pvpq:
    :param j1:
    :param j2:
    :param j3:
    :param j4:
    :param j5:
    :param j6:
    :param Va:
    :param Vm:
    :return:
    """

    npv = len(pv)
    npq = len(pq)

    # reassign the solution vector
    if npv:
        Va[pv] = x[j1:j2]
    if npq:
        Va[pq] = x[j3:j4]
        Vm[pq] = x[j5:j6]

    V = Vm * np.exp(1j * Va)  # voltage mismatch

    # right hand side
    g = F(V, Ybus, S, I, pq, pv)

    # jacobian
    gx = Jacobian(Ybus, V, I, pq, pvpq)

    # return the increment of x
    return spsolve(gx, g)


def runge_kutta_nr(Ybus, Sbus, V0, Ibus, pv, pq, tol, max_it=15, error_registry=None):
    """
    Solves the power flow using a the runge-kutta integration algorithm
    Args:
        Ybus: Admittance matrix
        Sbus: Array of nodal power injections
        V0: Array of nodal voltages (initial solution)
        Ibus: Array of nodal current injections
        pv: Array with the indices of the PV buses
        pq: Array with the indices of the PQ buses
        tol: Tolerance
        max_it: Maximum number of iterations
        error_registry: list of function evaluations
    Returns:
        Voltage solution, converged?, error, calculated power injections, elapsed time

    @Author: Santiago Peñate Vera
    """

    start = time.time()

    # initialize
    converged = False
    iter_ = 0
    V = V0.copy()

    # set up indexing for updating V
    pvpq = np.r_[pv, pq]
    npv = len(pv)
    npq = len(pq)

    # j1:j2 - V angle of pv buses
    j1 = 0
    j2 = npv
    # j3:j4 - V angle of pq buses
    j3 = j2
    j4 = j2 + npq
    # j5:j6 - V mag of pq buses
    j5 = j4
    j6 = j4 + npq

    # evaluate F(x0)
    Scalc = V * np.conj(Ybus * V - Ibus)
    mis = Scalc - Sbus  # compute the mismatch
    f = np.r_[mis[pv].real, mis[pq].real, mis[pq].imag]

    # check tolerance
    norm_f = np.linalg.norm(f, np.Inf)

    if norm_f < tol:
        converged = True

    if error_registry is not None:
        error_registry.append(norm_f)

    dt = 1.0

    # Compose x
    x = np.zeros(2 * npq + npv)
    Va = np.angle(V)
    Vm = np.abs(V)

    # do Newton iterations
    while not converged and iter_ < max_it:
        # update iteration counter
        iter_ += 1

        if npv:
            x[j1:j2] = Va[pv]
        if npq:
            x[j3:j4] = Va[pq]
            x[j5:j6] = Vm[pq]

        # Compute the Runge-Kutta steps
        k1 = fx(x,
                Ybus, Sbus, Ibus, pq, pv, pvpq, j1, j2, j3, j4, j5, j6, Va, Vm)

        k2 = fx(x + 0.5 * dt * k1,
                Ybus, Sbus, Ibus, pq, pv, pvpq, j1, j2, j3, j4, j5, j6, Va, Vm)

        k3 = fx(x + 0.5 * dt * k2,
                Ybus, Sbus, Ibus, pq, pv, pvpq, j1, j2, j3, j4, j5, j6, Va, Vm)

        k4 = fx(x + dt * k3,
                Ybus, Sbus, Ibus, pq, pv, pvpq, j1, j2, j3, j4, j5, j6, Va, Vm)

        x -= dt * (k1 + 2.0 * k2 + 2.0 * k3 + k4) / 6.0

        # re-assign the solution vector
        if npv:
            Va[pv] = x[j1:j2]
        if npq:
            Va[pq] = x[j3:j4]
            Vm[pq] = x[j5:j6]
        V = Vm * np.exp(1.0j * Va)  # voltage mismatch

        # evaluate F(x)
        Scalc = V * np.conj(Ybus * V - Ibus)
        mis = Scalc - Sbus  # complex power mismatch
        f = np.r_[mis[pv].real, mis[pq].real, mis[pq].imag]  # concatenate again

        # check for convergence
        norm_f = np.linalg.norm(f, np.Inf)

        if error_registry is not None:
            error_registry.append(norm_f)

        if norm_f > 0.01:
            dt = max(dt * 0.985, 0.75)
        else:
            dt = min(dt * 1.015, 0.75)

        if norm_f < tol:
            converged = True

    end = time.time()
    elapsed = end - start

    return V, converged, norm_f, Scalc, elapsed



########################################################################################################################
#  HELM
########################################################################################################################

@nb.njit("(c16[:])(c16[:, :], c16[:, :], i8, i8[:])")
def conv1(A, B, c, indices):
    """
    Performs the convolution of A* and B
    :param A: Coefficients matrix 1 (orders, buses)
    :param B: Coefficients matrix 2 (orders, buses)
    :param c: order of the coefficients
    :param indices: bus indices array
    :return: Array with the convolution for the buses given by "indices"
    """
    suma = np.zeros(len(indices), dtype=nb.complex128)
    for k in range(1, c + 1):
        for i, d in enumerate(indices):
            suma[i] += np.conj(A[k, d]) * B[c - k, d]
    return suma


@nb.njit("(c16[:])(c16[:, :], c16[:, :], i8, i8[:])")
def conv2(A, B, c, indices):
    """
    Performs the convolution of A and B
    :param A: Coefficients matrix 1 (orders, buses)
    :param B: Coefficients matrix 2 (orders, buses)
    :param c: order of the coefficients
    :param indices: bus indices array
    :return: Array with the convolution for the buses given by "indices"
    """
    suma = np.zeros(len(indices), dtype=nb.complex128)
    for k in range(1, c):
        for i, d in enumerate(indices):
            suma[i] += A[k, d] * B[c - 1 - k, d]
    return suma


@nb.njit("(c16[:])(c16[:, :], c16[:, :], i8, i8[:])")
def conv3(A, B, c, indices):
    """
    Performs the convolution of A and B*
    :param A: Coefficients matrix 1 (orders, buses)
    :param B: Coefficients matrix 2 (orders, buses)
    :param c: order of the coefficients
    :param indices: bus indices array
    :return: Array with the convolution for the buses given by "indices"
    """
    suma = np.zeros(len(indices), dtype=nb.complex128)
    for k in range(1, c):
        for i, d in enumerate(indices):
            suma[i] += A[k, d] * np.conj(B[c - k, d])
    return suma


def helm_coefficients_josep(Ybus, Yseries, V0, S0, Ysh0, pq, pv, sl, pqpv, tolerance=1e-6, max_coeff=30, verbose=False):
    """
    Holomorphic Embedding LoadFlow Method as formulated by Josep Fanals Batllori in 2020
    THis function just returns the coefficients for further usage in other routines
    :param Yseries: Admittance matrix of the series elements
    :param V0: vector of specified voltages
    :param S0: vector of specified power
    :param Ysh0: vector of shunt admittances (including the shunts of the branches)
    :param pq: list of pq nodes
    :param pv: list of pv nodes
    :param sl: list of slack nodes
    :param pqpv: sorted list of pq and pv nodes
    :param tolerance: target error (or tolerance)
    :param max_coeff: maximum number of coefficients
    :param verbose: print intermediate information
    :return: U, X, Q, iterations
    """

    npqpv = len(pqpv)
    npv = len(pv)
    nsl = len(sl)
    n = Yseries.shape[0]

    # --------------------------- PREPARING IMPLEMENTATION -------------------------------------------------------------
    U = np.zeros((max_coeff, npqpv), dtype=complex)  # voltages
    W = np.zeros((max_coeff, npqpv), dtype=complex)  # compute X=1/conj(U)
    Q = np.zeros((max_coeff, npqpv), dtype=complex)  # unknown reactive powers
    Vm0 = np.abs(V0)
    Vm2 = Vm0 * Vm0

    if n < 2:
        return U, W, Q, 0

    if verbose:
        print('Yseries')
        print(Yseries.toarray())
        df = pd.DataFrame(data=np.c_[Ysh0.imag, S0.real, S0.imag, Vm0],
                          columns=['Ysh', 'P0', 'Q0', 'V0'])
        print(df)

    Yred = Yseries[np.ix_(pqpv, pqpv)]  # admittance matrix without slack buses
    Yslack = -Yseries[np.ix_(pqpv, sl)]  # yes, it is the negative of this
    Yslack_vec = Yslack.sum(axis=1).A1
    G = np.real(Yred)  # real parts of Yij
    B = np.imag(Yred)  # imaginary parts of Yij
    P_red = S0.real[pqpv]
    Q_red = S0.imag[pqpv]
    Vslack = V0[sl]
    Ysh_red = Ysh0[pqpv]

    # indices 0 based in the internal scheme
    nsl_counted = np.zeros(n, dtype=int)
    compt = 0
    for i in range(n):
        if i in sl:
            compt += 1
        nsl_counted[i] = compt

    pq_ = pq - nsl_counted[pq]
    pv_ = pv - nsl_counted[pv]

    # .......................CALCULATION OF TERMS [0] ------------------------------------------------------------------

    U[0, :] = spsolve(Yred, Yslack_vec)
    W[0, :] = 1 / np.conj(U[0, :])

    # .......................CALCULATION OF TERMS [1] ------------------------------------------------------------------
    valor = np.zeros(npqpv, dtype=complex)

    # get the current injections that appear due to the slack buses reduction
    I_inj_slack = Yslack * Vslack

    valor[pq_] = I_inj_slack[pq_] - Yslack_vec[pq_] + (P_red[pq_] - Q_red[pq_] * 1j) * W[0, pq_] - U[0, pq_] * Ysh_red[pq_]
    valor[pv_] = I_inj_slack[pv_] - Yslack_vec[pv_] + P_red[pv_] * W[0, pv_] - U[0, pv_] * Ysh_red[pv_]

    # compose the right-hand side vector
    RHS = np.r_[valor.real,
                valor.imag,
                Vm2[pv] - (U[0, pv_] * U[0, pv_]).real]

    # Form the system matrix (MAT)
    Upv = U[0, pv_]
    Xpv = W[0, pv_]
    VRE = coo_matrix((2 * Upv.real, (np.arange(npv), pv_)), shape=(npv, npqpv)).tocsc()
    VIM = coo_matrix((2 * Upv.imag, (np.arange(npv), pv_)), shape=(npv, npqpv)).tocsc()
    XIM = coo_matrix((-Xpv.imag, (pv_, np.arange(npv))), shape=(npqpv, npv)).tocsc()
    XRE = coo_matrix((Xpv.real, (pv_, np.arange(npv))), shape=(npqpv, npv)).tocsc()
    EMPTY = csc_matrix((npv, npv))

    MAT = vs((hs((G,  -B,   XIM)),
              hs((B,   G,   XRE)),
              hs((VRE, VIM, EMPTY))), format='csc')

    if verbose:
        print('MAT')
        print(MAT.toarray())

    # factorize (only once)
    MAT_LU = factorized(MAT.tocsc())

    # solve
    LHS = MAT_LU(RHS)

    # update coefficients
    U[1, :] = LHS[:npqpv] + 1j * LHS[npqpv:2 * npqpv]
    Q[0, pv_] = LHS[2 * npqpv:]
    W[1, :] = -W[0, :] * np.conj(U[1, :]) / np.conj(U[0, :])

    # .......................CALCULATION OF TERMS [>=2] ----------------------------------------------------------------
    iter_ = 1
    range_pqpv = np.arange(npqpv, dtype=np.int64)
    V = V0.copy()
    c = 2
    converged = False
    norm_f = tolerance + 1.0  # any number that violates the convergence

    while c < max_coeff and not converged:  # c defines the current depth

        valor[pq_] = (P_red[pq_] - Q_red[pq_] * 1j) * W[c - 1, pq_] - U[c - 1, pq_] * Ysh_red[pq_]
        valor[pv_] = -1j * conv2(W, Q, c, pv_) - U[c - 1, pv_] * Ysh_red[pv_] + W[c - 1, pv_] * P_red[pv_]

        RHS = np.r_[valor.real,
                    valor.imag,
                    -conv3(U, U, c, pv_).real]

        LHS = MAT_LU(RHS)

        # update voltage coefficients
        U[c, :] = LHS[:npqpv] + 1j * LHS[npqpv:2 * npqpv]

        # update reactive power
        Q[c - 1, pv_] = LHS[2 * npqpv:]

        # update voltage inverse coefficients
        W[c, range_pqpv] = -conv1(U, W, c, range_pqpv) / np.conj(U[0, range_pqpv])

        # compute power mismatch
        if not np.mod(c, 2):  # check the mismatch every 4 iterations
            V[pqpv] = U.sum(axis=0)
            Scalc = V * np.conj(Ybus * V)
            dP = np.abs(S0[pqpv].real - Scalc[pqpv].real)
            dQ = np.abs(S0[pq].imag - Scalc[pq].imag)
            norm_f = np.linalg.norm(np.r_[dP, dQ], np.inf)  # same as max(abs())

            # check convergence
            converged = norm_f < tolerance
            print('mismatch check at c=', c)

        c += 1
        iter_ += 1

    return U, W, Q, iter_, norm_f


########################################################################################################################
#  MAIN
########################################################################################################################
if __name__ == "__main__":
    from GridCal.Engine import FileOpen, compile_snapshot_circuit
    from matplotlib import pyplot as plt
    import pandas as pd
    import os
    import time
    np.set_printoptions(linewidth=10000)
    pd.set_option('display.max_rows', 500)
    pd.set_option('display.max_columns', 500)
    pd.set_option('display.width', 1000)

    # fname = os.path.join('..', '..', '..', 'Grids_and_profiles', 'grids', 'IEEE 30 Bus with storage.xlsx')
    # fname = os.path.join('..', '..', '..', 'Grids_and_profiles', 'grids', 'Illinois200Bus.xlsx')
    # fname = os.path.join('..', '..', '..', 'Grids_and_profiles', 'grids', 'Pegase 2869.xlsx')
    # fname = os.path.join('..', '..', '..', 'Grids_and_profiles', 'grids', '1354 Pegase.xlsx')
    # fname = os.path.join('..', '..', '..', 'Grids_and_profiles', 'grids', 'IEEE 14.xlsx')
    # fname = '/home/santi/Documentos/Private_Grids/2026_INVIERNO_para Plexos_FINAL_9.raw'
    fname = '/home/santi/Documentos/Private_Grids/201902271115 caso TReal Israel.raw'

    grid = FileOpen(file_name=fname).open()
    nc = compile_snapshot_circuit(grid)
    islands = nc.split_into_islands(ignore_single_node_islands=True)
    circuit = islands[0]

    # declare figure
    fig = plt.figure(figsize=(12, 7))
    ax = fig.add_subplot(1, 1, 1)

    # circuit.Vbus = np.ones(len(circuit.Vbus), dtype=complex)

    print('Newton-Raphson-Line-search 3')
    for acc in [0.5]:  # [1e-6, 1e-5, 1e-4, 1e-3, 1e-2, 0.1]
        start_time = time.time()

        # print('Ybus')
        # print(circuit.Ybus.toarray())

        error_data1 = list()
        V1, converged_, err, S, el = NR_LS3(Ybus=circuit.Ybus,
                                            Sbus=circuit.Sbus,
                                            V0=circuit.Vbus,
                                            Ibus=circuit.Ibus,
                                            pv=circuit.pv,
                                            pq=circuit.pq,
                                            vd=circuit.vd,
                                            bus_installed_power=circuit.bus_installed_power / circuit.Sbase,
                                            tol=1e-10,
                                            max_it=50,
                                            acceleration_parameter=acc,
                                            error_registry=error_data1,
                                            correct_slack=True)

        print("--- %s seconds ---" % (time.time() - start_time))
        print('error: \t', err)
        ax.plot(error_data1, lw=2, label='NRLS 1:' + str(acc))

    # print('\nNewton-Raphson-Line-search 2')
    # for acc in [1e-6, 1e-5, 1e-4, 1e-3, 1e-2, 0.1]:
    #
    #     start_time = time.time()
    #     error_data2 = list()
    #     V1, converged_, err, S, el = NR_LS2(Ybus=circuit.Ybus,
    #                                         Sbus=circuit.Sbus,
    #                                         V0=circuit.Vbus,
    #                                         Ibus=circuit.Ibus,
    #                                         pv=circuit.pv,
    #                                         pq=circuit.pq,
    #                                         tol=1e-9,
    #                                         max_it=100,
    #                                         acceleration_parameter=acc,
    #                                         error_registry=error_data2)
    #
    #     print("--- %s seconds ---" % (time.time() - start_time))
    #     print('error: \t', err)
    #
    #     ax.plot(error_data2, lw=2, linestyle=':', label='NRLS 2: ' + str(acc))
    #
    # print('\nVanilla NR ---')
    # start_time = time.time()
    # error_data3 = list()
    # V1, converged_, err, S, el = NR(Ybus=circuit.Ybus,
    #                                 Sbus=circuit.Sbus,
    #                                 V0=circuit.Vbus,
    #                                 Ibus=circuit.Ibus,
    #                                 pv=circuit.pv,
    #                                 pq=circuit.pq,
    #                                 tol=1e-9,
    #                                 max_it=5,
    #                                 mu0=1.0,
    #                                 error_registry=error_data3)
    # ax.plot(error_data3, lw=2, linestyle='--', label='NR')
    #
    # print("--- %s seconds ---" % (time.time() - start_time))
    # print('error: \t', err)

    # print('\nRunge kutta ---')
    # start_time = time.time()
    # error_data3 = list()
    # V1, converged_, err, S, el = runge_kutta_nr(Ybus=circuit.Ybus,
    #                                             Sbus=circuit.Sbus,
    #                                             V0=circuit.Vbus,
    #                                             Ibus=circuit.Ibus,
    #                                             pv=circuit.pv,
    #                                             pq=circuit.pq,
    #                                             tol=1e-15,
    #                                             max_it=30,
    #                                             error_registry=error_data3)
    # ax.plot(error_data3, lw=2, linestyle='--', label='runge-kutta')
    #
    # print("--- %s seconds ---" % (time.time() - start_time))
    # print('error: \t', err)
    #
    # ax.set_yscale('log')
    # ax.set_xlabel('Evaluations of $f(x)$')
    # ax.set_ylabel('Error')
    # ax.legend()


    # print('\nHELM ---')
    # start_time = time.time()
    # n_coeff = 30
    #
    # # compute the series of coefficients
    # U, X, Q, iter_, norm_f = helm_coefficients_josep(Ybus=circuit.Ybus,
    #                                                  Yseries=circuit.Yseries,
    #                                                  V0=circuit.Vbus,
    #                                                  S0=circuit.Sbus,
    #                                                  Ysh0=circuit.Ysh_helm,
    #                                                  pv=circuit.pv,
    #                                                  pq=circuit.pq,
    #                                                  sl=circuit.ref,
    #                                                  pqpv=circuit.pqpv,
    #                                                  tolerance=1e-15,
    #                                                  max_coeff=n_coeff,
    #                                                  verbose=False)
    #
    # V = circuit.Vbus.copy()
    # Sbus = circuit.Sbus
    # Ybus = circuit.Ybus
    # pv = circuit.pv
    # pq = circuit.pq
    # error_data4 = list()
    # for c in range(n_coeff):
    #     V[circuit.pqpv] = U[:c, :].sum(axis=0)
    #     # evaluate F(x0)
    #     Scalc = V * np.conj(Ybus * V)
    #     mis = Scalc - Sbus  # compute the mismatch
    #     f = np.r_[mis[pv].real, mis[pq].real, mis[pq].imag]
    #
    #     # check tolerance
    #     norm_f = np.linalg.norm(f, np.Inf)
    #     error_data4.append(norm_f)
    #
    # ax.plot(error_data4, lw=2, linestyle='--', label='HELM')
    #
    # print("--- %s seconds ---" % (time.time() - start_time))
    # print('error: \t', err)
    #
    # ax.set_yscale('log')
    # ax.set_xlabel('Evaluations of $f(x)$')
    # ax.set_ylabel('Error')
    # ax.legend()
    # check against the standard NR power flow used in GridCal
    # print('\nNR implemented in GridCal ---')
    # options = PowerFlowOptions(SolverType.NR, verbose=False, tolerance=1e-9, control_q=False)
    # power_flow = PowerFlowDriver(grid, options)
    #
    # start_time = time.time()
    # power_flow.run()
    # print("--- %s seconds ---" % (time.time() - start_time))
    # vnr = power_flow.results.voltage
    #
    # print('error: \t', power_flow.results.error)
    #
    # # check
    # data = np.c_[np.abs(V1), angle(V1), np.abs(vnr), angle(vnr),  np.abs(V1 - vnr)]
    # cols = ['|V|', 'angle', '|V| benchmark NR', 'angle benchmark NR', 'Diff']
    # df = pd.DataFrame(data, columns=cols)

    print()
    # print(df)

    # plt.show()



"""
N phase calculation engine
(c) Santiago Peñate Vera, 2018
"""

import numpy as np
from scipy.sparse.linalg import spsolve
from scipy.sparse import csc_matrix, hstack, vstack
from scipy.sparse.linalg import factorized
from scipy.sparse.linalg import splu


def gauss_seidel_power_flow(Vbus, Sbus, Ibus, Ybus,
                            P0, Q0, exp_p, exp_q, V0,
                            A, B, C,
                            pq, pv, tol, max_iter, verbose=False):
    """
    Gauss-Seidel power flow
    :param Vbus: Bus voltage complex vector
    :param Sbus: Bus complex power injections vector
    :param Ibus: Bus complex current injections vector
    :param Ybus: Nodal admittance matrix (complex and sparse)
    :param P0: Exponential load parameter P0
    :param Q0: Exponential load parameter Q0
    :param exp_p: Exponential load parameter exp_p
    :param exp_q: Exponential load parameter exp_q
    :param V0: Exponential load parameter V0
    :param A: Polynomial load parameter A
    :param B: Polynomial load parameter B
    :param C: Polynomial load parameter C
    :param pq: list of pq marked nodes
    :param pv: list of pv marked nodes
    :param tol: tolerance of the solution
    :param max_iter: Maximum number of iterations
    :return: Voltage vector (solution), converged?, power error
    """

    factor = 0.9

    V = Vbus.copy()
    Vm = np.abs(V)

    # compute error
    mis = V * np.conj(Ybus * V - Ibus) - Sbus
    F = np.r_[mis[pv].real, mis[pq].real, mis[pq].imag]
    error = np.linalg.norm(F, np.Inf)

    # check convergence
    converged = error < tol

    # Gauss-Seidel
    iter_ = 0
    while not converged and iter_ < max_iter:

        # compute the exponential load model injection
        Vm = np.abs(V)
        Pexp = P0 / (np.power(V0, exp_p)) * np.power(Vm, exp_p)
        Qexp = Q0 / (np.power(V0, exp_q)) * np.power(Vm, exp_q)
        Sexp = Pexp + 1j * Qexp

        # compute the polynomial load model
        Spoly = A + B * Vm + C * np.power(Vm, 2.0)

        for k in pq:
            V[k] += factor * (np.conj((Sbus[k] - Sexp[k] - Spoly[k]) / V[k] + Ibus[k]) - Ybus[k, :] * V) / Ybus[k, k]  # compute the voltage

        for k in pv:
            # get the reactive power
            Q = (V[k] * np.conj(Ybus[k, :] * V - Ibus[k])).imag
            # compose the new complex power
            Sbus[k] = Sbus[k].real + 1j * Q
            # compute the voltage
            V[k] += factor * (np.conj((Sbus[k] - Sexp[k] - Spoly[k]) / V[k]) - Ybus[k, :] * V) / Ybus[k, k]
            # correct the voltage with the specified module of the voltage
            V[k] *= Vm[k] / np.abs(V[k])

        # compute error
        Scalc = V * np.conj(Ybus * V - Ibus)  # computed nodal power
        mis = Scalc - (Sbus - Sexp - Spoly)  # power mismatch
        F = np.r_[mis[pv].real, mis[pq].real, mis[pq].imag]  # array of particular mismatch values
        error = np.linalg.norm(F, np.Inf)  # infinite norm of the mismatch vector

        # check convergence
        converged = error < tol

        iter_ += 1
        if verbose:
            print('V: iter:', iter_, 'err:', error)
            print(np.abs(V))

    return V, converged, error


def jacobian(Ybus, Vbus, Ibus, pq, pvpq):
    """
    Computes the system Jacobian matrix
    Args:
        Ybus: Admittance matrix
        Vbus: Array of nodal voltages
        Ibus: Array of nodal current injections
        pq: Array with the indices of the PQ buses
        pvpq: Array with the indices of the PV and PQ buses in that precise order
    Returns:
        The system Jacobian matrix
    """
    ib = range(len(Vbus))
    Ibus = Ybus * Vbus - Ibus

    diagV = csc_matrix((Vbus, (ib, ib)))
    diagIbus = csc_matrix((Ibus, (ib, ib)))
    diagVnorm = csc_matrix((Vbus / np.abs(Vbus), (ib, ib)))

    dS_dVm = diagV * np.conj(Ybus * diagVnorm) + np.conj(diagIbus) * diagVnorm
    dS_dVa = 1j * diagV * np.conj(diagIbus - Ybus * diagV)

    J11 = dS_dVa[np.array([pvpq]).T, pvpq].real
    J12 = dS_dVm[np.array([pvpq]).T, pq].real
    J21 = dS_dVa[np.array([pq]).T, pvpq].imag
    J22 = dS_dVm[np.array([pq]).T, pq].imag

    J = vstack([hstack([J11, J12]),
                hstack([J21, J22])], format="csr")

    return J


def newton_raphson_power_flow(Vbus, Sbus, Ibus, Ybus,
                              P0, Q0, exp_p, exp_q, V0,
                              A, B, C,
                              pq, pv, tol, max_iter, verbose=False):
    """
    Solves the power flow using a full Newton's method with the Iwamoto optimal step factor.
    Args:
        Vbus: Array of nodal voltages (initial solution)
        Sbus: Array of nodal power injections
        Ibus: Array of nodal current injections
        Ybus: Admittance matrix
        P0: Exponential load parameter P0
        Q0: Exponential load parameter Q0
        exp_p: Exponential load parameter exp_p
        exp_q: Exponential load parameter exp_q
        V0: Exponential load parameter V0
        A: Polynomial load parameter A
        B: Polynomial load parameter B
        C: Polynomial load parameter C
        pv: Array with the indices of the PV buses
        pq: Array with the indices of the PQ buses
        tol: Tolerance
        max_it: Maximum number of iterations
        robust: Boolean variable for the use of the Iwamoto optimal step factor.
    Returns:
        Voltage solution, converged?, error, calculated power injections

    @author: Ray Zimmerman (PSERC Cornell)
    @Author: Santiago Penate Vera
    """
    # initialize
    converged = 0
    iter_ = 0
    V = Vbus
    Va = np.angle(V)
    Vm = np.abs(V)
    dVa = np.zeros_like(Va)
    dVm = np.zeros_like(Vm)

    # set up indexing for updating V
    pvpq = np.r_[pv, pq]
    npv = len(pv)
    npq = len(pq)

    # j1:j2 - V angle of pv buses
    j1 = 0
    j2 = npv
    # j3:j4 - V angle of pq buses
    j3 = j2
    j4 = j2 + npq
    # j5:j6 - V mag of pq buses
    j5 = j4
    j6 = j4 + npq

    # evaluate F(x0)
    Scalc = V * np.conj(Ybus * V - Ibus)

    # compute the exponential load model injection
    Vabs = np.abs(V)
    Pexp = P0 / (np.power(V0, exp_p)) * np.power(Vabs, exp_p)
    Qexp = Q0 / (np.power(V0, exp_q)) * np.power(Vabs, exp_q)
    Sexp = Pexp + 1j * Qexp

    # compute the polynomial load model
    Spoly = A + B * Vabs + C * np.power(Vabs, 2.0)

    mis = Scalc - Sbus - Sexp - Spoly  # compute the mismatch
    F = np.r_[mis[pv].real,
              mis[pq].real,
              mis[pq].imag]

    # check tolerance
    error = np.linalg.norm(F, np.Inf)

    if error < tol:
        converged = 1

    # do Newton iterations
    while not converged and iter_ < max_iter:
        # update iteration counter
        iter_ += 1

        # evaluate Jacobian
        J = jacobian(Ybus, V, Ibus, pq, pvpq)

        # compute update step
        dx = spsolve(J, F)

        # reassign the solution vector
        if npv:
            dVa[pv] = dx[j1:j2]
        if npq:
            dVa[pq] = dx[j3:j4]
            dVm[pq] = dx[j5:j6]
        # dV = dVm * np.exp(1j * dVa)  # voltage mismatch

        # update voltage
        Vm -= dVm
        Va -= dVa

        V = Vm * np.exp(1j * Va)

        Vm = np.abs(V)  # update Vm and Va again in case
        Va = np.angle(V)  # we wrapped around with a negative Vm

        # evaluate F(x)
        Scalc = V * np.conj(Ybus * V - Ibus)

        # compute the exponential load model injection
        Pexp = P0 / (np.power(V0, exp_p)) * np.power(Vm, exp_p)
        Qexp = Q0 / (np.power(V0, exp_q)) * np.power(Vm, exp_q)
        Sexp = Pexp + 1j * Qexp

        # compute the polynomial load model
        Spoly = A + B * Vm + C * np.power(Vm, 2.0)

        mis = Scalc - Sbus - Sexp - Spoly  # complex power mismatch
        F = np.r_[mis[pv].real, mis[pq].real, mis[pq].imag]  # concatenate again

        # check for convergence
        error = np.linalg.norm(F, np.Inf)

        if error < tol:
            converged = 1

        if verbose:
            print('V: iter:', iter_, 'err:', error)
            print(np.abs(V))

    return V, converged, error


def levenberg_marquardt_power_flow(Vbus, Sbus, Ibus, Ybus,
                                   P0, Q0, exp_p, exp_q, V0,
                                   A, B, C,
                                   pq, pv, tol, max_iter=50, verbose=False):
    """
    Solves the power flow problem by the Levenberg-Marquardt power flow algorithm.
    It is usually better than Newton-Raphson, but it takes an order of magnitude more time to converge.
    Args:
        Vbus: Array of nodal voltages (initial solution)
        Sbus: Array of nodal power injections
        Ibus: Array of nodal current injections
        Ybus: Admittance matrix
        P0: Exponential load parameter P0
        Q0: Exponential load parameter Q0
        exp_p: Exponential load parameter exp_p
        exp_q: Exponential load parameter exp_q
        V0: Exponential load parameter V0
        A: Polynomial load parameter A
        B: Polynomial load parameter B
        C: Polynomial load parameter C
        pv: Array with the indices of the PV buses
        pq: Array with the indices of the PQ buses
        tol: Tolerance
        max_iter: Maximum number of iterations
    Returns:
        Voltage solution, converged?, error, calculated power injections

    @Author: Santiago Peñate Vera
    """
    # initialize
    V = Vbus
    Va = np.angle(V)
    Vm = np.abs(V)
    dVa = np.zeros_like(Va)
    dVm = np.zeros_like(Vm)
    # set up indexing for updating V
    pvpq = np.r_[pv, pq]
    npv = len(pv)
    npq = len(pq)

    # j1:j2 - V angle of pv buses
    j1 = 0
    j2 = npv
    # j3:j4 - V angle of pq buses
    j3 = j2
    j4 = j2 + npq
    # j5:j6 - V mag of pq buses
    j5 = j4
    j6 = j4 + npq

    update_jacobian = True
    converged = False
    iter_ = 0
    nu = 2.0
    lbmda = 0
    f_prev = 1e9  # very large number
    nn = 2 * npq + npv
    ii = np.linspace(0, nn-1, nn)
    Idn = csc_matrix((np.ones(nn), (ii, ii)), shape=(nn, nn))  # csr_matrix identity

    while not converged and iter_ < max_iter:

        # evaluate Jacobian
        if update_jacobian:
            H = jacobian(Ybus, V, Ibus, pq, pvpq)

        # evaluate the solution error F(x0)
        Scalc = V * np.conj(Ybus * V - Ibus)
        # compute the exponential load model injection
        Vm = np.abs(V)
        Pexp = P0 / (np.power(V0, exp_p)) * np.power(Vm, exp_p)
        Qexp = Q0 / (np.power(V0, exp_q)) * np.power(Vm, exp_q)
        Sexp = Pexp + 1j * Qexp

        # compute the polynomial load model
        Spoly = A + B * Vm + C * np.power(Vm, 2.0)

        mis = Scalc - Sbus - Sexp - Spoly  # complex power mismatch
        dz = np.r_[mis[pv].real, mis[pq].real, mis[pq].imag]  # mismatch in the Jacobian order

        # system matrix
        # H1 = H^t
        H1 = H.transpose().tocsr()

        # H2 = H1·H
        H2 = H1.dot(H)

        # set first value of lmbda
        if iter_ == 0:
            lbmda = 1e-3 * H2.diagonal().max()

        # compute system matrix A = H^T·H - lambda·I
        Amat = H2 + lbmda * Idn

        # right hand side
        # H^t·dz
        rhs = H1.dot(dz)

        # Solve the increment
        dx = spsolve(Amat, rhs)

        # objective function to minimize
        f = 0.5 * dz.dot(dz)

        # decision function
        val = dx.dot(lbmda * dx + rhs)
        if val > 0.0:
            rho = (f_prev - f) / (0.5 * val)
        else:
            rho = -1.0

        # lambda update
        if rho >= 0:
            update_jacobian = True
            lbmda *= max([1.0 / 3.0, 1 - (2 * rho - 1) ** 3])
            nu = 2.0

            # reassign the solution vector
            if npv:
                dVa[pv] = dx[j1:j2]
            if npq:
                dVa[pq] = dx[j3:j4]
                dVm[pq] = dx[j5:j6]

            Vm -= dVm
            Va -= dVa
            # update Vm and Va again in case we wrapped around with a negative Vm
            V = Vm * np.exp(1j * Va)
            Vm = np.abs(V)
            Va = np.angle(V)
        else:
            update_jacobian = False
            lbmda *= nu
            nu *= 2.0

        # check convergence
        f_prev = f

        # evaluate F(x)
        Scalc = V * np.conj(Ybus * V - Ibus)

        # compute the exponential load model injection
        Pexp = P0 / (np.power(V0, exp_p)) * np.power(Vm, exp_p)
        Qexp = Q0 / (np.power(V0, exp_q)) * np.power(Vm, exp_q)
        Sexp = Pexp + 1j * Qexp

        # compute the polynomial load model
        Spoly = A + B * Vm + C * np.power(Vm, 2.0)

        mis = Scalc - Sbus - Sexp - Spoly  # complex power mismatch
        mismatch = np.r_[mis[pv].real, mis[pq].real, mis[pq].imag]  # concatenate again

        # check for convergence
        error = np.linalg.norm(mismatch, np.Inf)

        if error < tol:
            converged = 1

        # update iteration counter
        iter_ += 1

    return V, converged, error


def linearized_dc_power_flow(Ybus, Sbus, Ibus, V0, ref, pq, pv):
    """
    Solves a DC power flow.
    :param Ybus: Normal circuit admittance matrix
    :param Sbus: Complex power injections at all the nodes
    :param Ibus: Complex current injections at all the nodes
    :param V0: Array of complex seed voltage (it contains the ref voltages)
    :param ref: array of the indices of the slack nodes
    :param pvpq: array of the indices of the non-slack nodes
    :param pq: array of the indices of the pq nodes
    :param pv: array of the indices of the pv nodes
    :return:
        Complex voltage solution
        Converged: Always true
        Solution error
        Computed power injections given the found solution
    """

    pvpq = np.r_[pv, pq].astype(int)

    # Decompose the voltage in angle and magnitude
    Va_ref = np.angle(V0[ref])  # we only need the angles at the slack nodes
    Vm = np.abs(V0)

    # initialize result vector
    Va = np.empty(len(V0))

    # reconvert the pqpv vector to a matrix so that we can call numpy directly with it
    pvpq_ = np.matrix(pvpq)

    # Compile the reduced imaginary impedance matrix
    Bpqpv = Ybus.imag[pvpq_.T, pvpq_]
    Bref = Ybus.imag[pvpq_.T, ref]

    # compose the reduced power injections
    # Since we have removed the slack nodes, we must account their influence as injections Bref * Va_ref
    Pinj = Sbus[pvpq].real + (- Bref * Va_ref + Ibus[pvpq].real) * Vm[pvpq]

    # update angles for non-reference buses
    Va[pvpq] = spsolve(Bpqpv, Pinj)
    Va[ref] = Va_ref

    # re assemble the voltage
    V = Vm * np.exp(1j * Va)

    # compute the calculated power injection and the error of the voltage solution
    Scalc = V * np.conj(Ybus * V - Ibus)

    # compute the power mismatch between the specified power Sbus and the calculated power Scalc
    mis = Scalc - Sbus  # complex power mismatch
    F = np.r_[mis[pv].real, mis[pq].real, mis[pq].imag]  # concatenate again

    # check for convergence
    normF = np.linalg.norm(F, np.Inf)

    return V, True, normF


def linear_ac_power_flow(Ybus, Yseries, Sbus, Ibus, Vbus, pq, pv):
    """
    Linearized AC Load Flow

    form the article:

    Linearized AC Load Flow Applied to Analysis in Electric Power Systems
        by: P. Rossoni, W. M da Rosa and E. A. Belati
    Args:
        Ybus: Admittance matrix
        Yseries: Admittance matrix of the series elements
        Sbus: Power injections vector of all the nodes
        Vbus: Set voltages of all the nodes (used for the slack and PV nodes)
        pq: list of indices of the pq nodes
        pv: list of indices of the pv nodes

    Returns: Voltage vector, converged?, error, calculated power and elapsed time
    """
    pvpq = np.r_[pv, pq].astype(int)
    npq = len(pq)
    npv = len(pv)

    # compose the system matrix
    # G = Y.real
    # B = Y.imag
    # Gp = Ys.real
    # Bp = Ys.imag

    # A11 = -Yseries.imag[pvpq, :][:, pvpq]
    # A12 = Ybus.real[pvpq, :][:, pq]
    # A21 = -Yseries.real[pq, :][:, pvpq]
    # A22 = -Ybus.imag[pq, :][:, pq]

    A11 = -Yseries.imag[np.ix_(pvpq, pvpq)]
    A12 = Ybus.real[np.ix_(pvpq, pq)]
    A21 = -Yseries.real[np.ix_(pq, pvpq)]
    A22 = -Ybus.imag[np.ix_(pq, pq)]

    Asys = vstack([hstack([A11, A12]),
                   hstack([A21, A22])], format="csc")

    # compose the right hand side (power vectors)
    rhs = np.r_[Sbus.real[pvpq], Sbus.imag[pq]]

    # solve the linear system
    x = factorized(Asys)(rhs)

    # compose the results vector
    voltages_vector = Vbus.copy()

    #  set the pv voltages
    va_pv = x[0:npv]
    vm_pv = np.abs(Vbus[pv])
    voltages_vector[pv] = vm_pv * np.exp(1j * va_pv)

    # set the PQ voltages
    va_pq = x[npv:npv+npq]
    vm_pq = np.ones(npq) + x[npv+npq::]
    voltages_vector[pq] = vm_pq * np.exp(1j * va_pq)

    # Calculate the error and check the convergence
    s_calc = voltages_vector * np.conj(Ybus * voltages_vector)
    # complex power mismatch
    power_mismatch = s_calc - Sbus
    # concatenate error by type
    mismatch = np.r_[power_mismatch[pv].real, power_mismatch[pq].real, power_mismatch[pq].imag]

    # check for convergence
    norm_f = np.linalg.norm(mismatch, np.Inf)

    return voltages_vector, True, norm_f


def z_matrix_power_flow(Vbus, Sbus, Ibus, Ybus,
                        P0, Q0, exp_p, exp_q, V0,
                        A, B, C,
                        pq, pv, ref, tol, max_iter, verbose=False):
    """
    Solves the power flow using a full Newton's method with the Iwamoto optimal step factor.
    Args:
        Vbus: Array of nodal voltages (initial solution)
        Sbus: Array of nodal power injections
        Ibus: Array of nodal current injections
        Ybus: Admittance matrix
        P0: Exponential load parameter P0
        Q0: Exponential load parameter Q0
        exp_p: Exponential load parameter exp_p
        exp_q: Exponential load parameter exp_q
        V0: Exponential load parameter V0
        A: Polynomial load parameter A
        B: Polynomial load parameter B
        C: Polynomial load parameter C
        pv: Array with the indices of the PV buses
        pq: Array with the indices of the PQ buses
        tol: Tolerance
        max_it: Maximum number of iterations
        robust: Boolean variable for the use of the Iwamoto optimal step factor.
    Returns:
        Voltage solution, converged?, error, calculated power injections

    @Author: Santiago Penate Vera
    """

    """
    From the paper:  Three-phase distribution network fast-decoupled power flow solutions
    
    |G  -B||de|   |dIr|
    |     ||  | = |   |
    |B   G||df|   |dIi|
    
    """

    pqpv = np.r_[pq, pv].astype(int)
    npq = len(pq)
    npv = len(pv)

    # reduced impedance matrix
    Zred = factorized(Ybus[pqpv, :][:, pqpv])

    # slack currents
    Ivd = Ybus[pqpv, :][:, ref].dot(Vbus[ref])
    print('Ivd', np.vstack(Ivd))

    # slack voltages influence
    Ck = Zred(Ivd)
    print('Ck', np.vstack(Ck))

    # make a copy of the voltage for convergence control
    Vprev = Vbus[pqpv].copy()

    # Voltage module in the pv nodes
    Vpv = np.abs(Vbus[pv])

    # admittance matrix to compute the reactive power
    Ybus_pv = Ybus[pv, :][:, pv]

    # approximate the currents with the current voltage solution
    Ik = np.conj(Sbus[pqpv] / Vprev) + Ibus[pqpv]
    print('Sred', np.vstack(Sbus[pqpv]))
    print('Ik', np.vstack(Ik))

    # compute the new voltage solution
    Vk = Zred(Ik) - Ck
    print('Vk', np.vstack(Vk))

    # compute the voltage solution maximum difference
    diff = np.max(np.abs(Vprev - Vk))

    iter = 1
    while diff > tol and iter < max_iter:
        # make a copy of the voltage for convergence control
        Vprev = Vk

        # approximate the currents with the current voltage solution
        Ik = np.conj(Sbus[pqpv] / Vprev) + Ibus[pqpv]

        # compute the new voltage solution
        Vk = Zred(Ik) - Ck
        print(iter, 'Vk', Vk)
        print()
        # tune PV nodes
        #  ****** USE A reduced pv, pv, pqpv mapping!
        # Vk[pv] *= Vpv / abs(Vk[pv])
        # Qpv = (Vk * conj(Ybus[pv, :][:, pv].dot(Vk) - Ibus))[pv].imag
        # Sbus[pv] = Sbus[pv].real + 1j * Qpv

        # compute the voltage solution maximum difference
        diff = np.max(np.abs(Vprev - Vk))

        # Assign the reduced voltage solution to the complete voltage solution
        # voltage = Vbus.copy()  # the slack voltages are kept
        # voltage[pqpv] = Vk
        # compute the power mismatch: this is the true equation solution check
        # Scalc = voltage * conj(Ybus * voltage - Ibus)
        # mis = Scalc - Sbus  # complex power mismatch
        # diff = linalg.norm(r_[mis[pv].real, mis[pq].real, mis[pq].imag], Inf)

        iter += 1

    # Assign the reduced voltage solution to the complete voltage solution
    voltage = Vbus.copy()  # the slack voltages are kept
    voltage[pqpv] = Vk

    print(iter, 'voltage:\n', np.vstack(voltage))
    print()

    # compute the power mismatch: this is the true equation solution check
    Scalc = voltage * np.conj(Ybus * voltage - Ibus)
    mis = Scalc - Sbus  # complex power mismatch
    norm_f = np.linalg.norm(np.r_[mis[pv].real, mis[pq].real, mis[pq].imag], np.Inf)

    return voltage, True, norm_f


def gauss_raphson_power_flow(Vbus, Sbus, Ibus, Ybus,
                             P0, Q0, exp_p, exp_q, V0,
                             A, B, C,
                             pq, pv, ref, tol, max_iter, verbose=False):
    """
    Solves the power flow using a full Newton's method with the Iwamoto optimal step factor.
    Args:
        Vbus: Array of nodal voltages (initial solution)
        Sbus: Array of nodal power injections
        Ibus: Array of nodal current injections
        Ybus: Admittance matrix
        P0: Exponential load parameter P0
        Q0: Exponential load parameter Q0
        exp_p: Exponential load parameter exp_p
        exp_q: Exponential load parameter exp_q
        V0: Exponential load parameter V0
        A: Polynomial load parameter A
        B: Polynomial load parameter B
        C: Polynomial load parameter C
        pv: Array with the indices of the PV buses
        pq: Array with the indices of the PQ buses
        tol: Tolerance
        max_it: Maximum number of iterations
        robust: Boolean variable for the use of the Iwamoto optimal step factor.
    Returns:
        Voltage solution, converged?, error, calculated power injections

    @Author: Santiago Penate Vera
    """

    """
    From the paper:  Three-phase distribution network fast-decoupled power flow solutions

    |G  -B||de|   |dIr|
    |     ||  | = |   |
    |B   G||df|   |dIi|

    """

    pvpq = np.r_[pv, pq].astype(int)
    npq = len(pq)
    npv = len(pv)
    npvpq = npv + npq

    V = Vbus.copy()
    G = Ybus.real[pvpq, :][:, pvpq]
    B = Ybus.imag[pvpq, :][:, pvpq]

    J = vstack([hstack([G, B]),
                hstack([B, G])], format="csc")

    for i in range(1):
        # complex current mismatch
        Isp = np.conj(Sbus / V)
        Icalc = Ybus * V
        dI = Isp - Icalc
        I = np.hstack((dI.real[pvpq], dI.imag[pvpq]))

        # solve the rectangular voltage increments
        # x = spsolve(J, I)
        # x = J * I

        # assign the voltage values
        dVr = spsolve(G, dI.real[pvpq])
        dVi = spsolve(G, dI.imag[pvpq])

        V[pvpq] -= dVr + 1j * dVi
        print("V: ", np.abs(V))

    # Calculate the error and check the convergence
    s_calc = V * np.conj(Ybus * V)
    # complex power mismatch
    power_mismatch = s_calc - Sbus
    # concatenate error by type
    mismatch = np.r_[power_mismatch[pv].real, power_mismatch[pq].real, power_mismatch[pq].imag]

    # check for convergence
    norm_f = np.linalg.norm(mismatch, np.Inf)

    return V, True, norm_f


def fast_decoupled_power_flow(Vbus, Sbus, Ibus, Ybus, B1, B2, pq, pv, pqpv, tol=1e-9, max_iter=100):
    """

    :param Vbus:
    :param Sbus:
    :param Ibus:
    :param Ybus:
    :param B1:
    :param B2:
    :param pq:
    :param pv:
    :param pqpv:
    :param tol:
    :param max_iter:
    :return:
    """

    # set voltage vector for the iterations
    voltage = Vbus.copy()
    Va = np.angle(voltage)
    Vm = np.abs(voltage)

    # Factorize B1 and B2
    J1 = splu(B1[pqpv, :][:, pqpv])
    J2 = splu(B2[pq, :][:, pq])

    # evaluate initial mismatch
    Scalc = voltage * np.conj(Ybus * voltage - Ibus)
    mis = Scalc - Sbus  # complex power mismatch
    incP = mis[pqpv].real
    incQ = mis[pq].imag
    normP = np.norm(incP, np.Inf)
    normQ = np.norm(incQ, np.Inf)
    if normP < tol and normQ < tol:
        converged = True
    else:
        converged = False

    # iterate
    iter_ = 0
    while not converged and iter_ < max_iter:

        iter_ += 1

        # solve voltage angles
        dVa = -J1.solve(incP)

        # update voltage
        Va[pqpv] = Va[pqpv] + dVa
        voltage = Vm * np.exp(1j * Va)

        # evaluate mismatch
        Scalc = voltage * np.conj(Ybus * voltage - Ibus)
        mis = Scalc - Sbus  # complex power mismatch
        incP = mis[pqpv].real
        incQ = mis[pq].imag
        normP = np.norm(incP, np.Inf)
        normQ = np.norm(incQ, np.Inf)

        if normP < tol and normQ < tol:
            converged = True

        else:
            # Solve voltage modules
            dVm = -J2.solve(incQ)

            # update voltage
            Vm[pq] = Vm[pq] + dVm
            voltage = Vm * np.exp(1j * Va)

            # evaluate mismatch
            Scalc = voltage * np.conj(Ybus * voltage - Ibus)
            mis = Scalc - Sbus  # complex power mismatch
            incP = mis[pqpv].real
            incQ = mis[pq].imag
            normP = np.norm(incP, np.Inf)
            normQ = np.norm(incQ, np.Inf)

            if normP < tol and normQ < tol:
                converged = True

    # evaluate F(x)
    Scalc = voltage * np.conj(Ybus * voltage - Ibus)
    mis = Scalc - Sbus  # complex power mismatch
    F = np.r_[mis[pv].real, mis[pq].real, mis[pq].imag]  # concatenate again

    # check for convergence
    normF = np.norm(F, np.Inf)

    return voltage, converged, normF



import numpy as np
from numpy import where, zeros, ones, mod, conj, array, dot, complex128, linspace, angle  # , complex256
from scipy.linalg import solve
from scipy.sparse import dia_matrix, coo_matrix, csc_matrix, hstack as sp_hstack, vstack as sp_vstack
from scipy.sparse.linalg import factorized
# Set the complex precision to use
complex_type = complex128


def pade_approximation(n, d, an, s=1):
    """
    Computes the n/2 pade approximant of the series an at the approximation
    point s

    Arguments:
        an: coefficient series
        n:  order of the series
        d:  bus index
        s: point of approximation

    Returns:
        pade approximation at s
    """
    nn = int(n / 2)
    L = nn
    M = nn

    # formation of the linear system right hand side
    rhs = an[L + 1:L + M + 1, d]

    # formation of the coefficients matrix
    C = zeros((L, M), dtype=complex_type)
    for i in range(L):
        k = i + 1
        C[i, :] = an[L - M + k:L + k, d]

    # Obtaining of the b coefficients for orders greater than 0
    b = solve(C, -rhs)  # bn to b1
    b = r_[1, b[::-1]]  # b0 = 1

    # Obtaining of the coefficients 'a'
    a = zeros(L + 1, dtype=complex_type)
    a[0] = an[0, d]
    for i in range(L):
        val = complex_type(0)
        k = i + 1
        for j in range(k + 1):
            val += an[k - j, d] * b[j]
        a[i + 1] = val

    # evaluation of the function for the value 's'
    p = complex_type(0)
    q = complex_type(0)
    for i in range(L + 1):
        p += a[i] * s ** i
        q += b[i] * s ** i

    return p / q, a, b


def make_A2(Y_series, Y_shunt, pq, pv, pqpv, types):
    """

    Args:
        Y_series: series admittances matrix
        Y_shunt: shunt admittances vector
        pq:
        pv:
    Returns:

    """

    # create system matrix A of the model 4 of the Wallace article
    NPQ = len(pq)
    NPV = len(pv)
    NPQPV = NPQ + NPV

    dij_y = dia_matrix((Y_shunt[pqpv], zeros(1)), shape=(NPQPV, NPQPV)).tocsc()
    dij_pv = coo_matrix((ones(NPV) * 2, (linspace(0, NPV - 1, NPV), pv)), shape=(NPV, NPQPV)).tocsc()

    G = Y_series.real[pqpv, :][:, pqpv]
    B = Y_series.imag[pqpv, :][:, pqpv]

    Gpq = csc_matrix((NPQ, NPQPV))
    Bpq = csc_matrix((NPQ, NPQPV))
    dij_y_pq = csc_matrix((NPQ, NPQPV), dtype=complex_type)
    ii = 0
    for i, ti in enumerate(types):
        if ti == 1:
            jj = 0
            for j, tj in enumerate(types):
                if tj == 1:
                    Gpq[ii, jj] = Y_series.real[i, j]
                    Bpq[ii, jj] = Y_series.imag[i, j]
                    if ii == jj:
                        dij_y_pq[ii, jj] = Y_shunt[i]
                if tj == 1 or tj == 2:
                    jj += 1
            ii += 1

    print('G:\n', Y_series.real.toarray())
    print('B:\n', Y_series.imag.toarray())
    print('dij_y_pq:\n', dij_y_pq.toarray())

    A1 = G + dij_y.real
    print('\nA1:\n', G.toarray(), '\n+\n', dij_y.real.toarray())

    A2 = B - dij_y.imag
    print('\nA2:\n', B.toarray(), '\n-\n', dij_y.imag.toarray())

    APQ3 = Bpq - dij_y_pq.imag
    print('\nA_PQ3:\n', Bpq.toarray(), '\n-\n', dij_y_pq.imag.toarray())

    APQ4 = Gpq + dij_y_pq.real
    print('\nA_PQ4:\n', Gpq.toarray(), '\n+\n', dij_y_pq.real.toarray())

    APV3 = dij_pv
    print('\nA_PV3:\n', dij_pv.toarray())

    APV4 = csc_matrix((NPV, NPQPV))
    print('\nA_PV4:\n', APV4.toarray())

    A = sp_vstack((
        sp_hstack((A1, A2)),
        sp_hstack((APQ3, APQ4)),
        sp_hstack((APV3, APV4))
    )).tocsc()

    return A, NPQPV


def make_A(Y_series, Y_shunt, pq, pv, pqpv, types):

    # create system matrix A of the model 4 of the Wallace article
    N = len(Y_shunt)
    NPQ = len(pq)
    NPV = len(pv)
    NPQPV = NPQ + NPV

    dij_y = dia_matrix((Y_shunt, zeros(1)), shape=(N, N)).tocsc()
    dij = dia_matrix((ones(N), zeros(1)), shape=(N, N)).tocsc()
    G = Y_series.real
    B = Y_series.imag

    A1 = (G + dij_y.real)[pqpv, :][:, pqpv]
    A2 = (B - dij_y.imag)[pqpv, :][:, pqpv]

    M = (B - dij_y.imag)
    M[:, pv] = zeros((N, 1))
    APQ3 = M[pq, :][:, pqpv]

    M = (G + dij_y.imag)
    M[:, pv] = zeros((N, 1))
    APQ4 = M[pq, :][:, pqpv]

    APV3 = (2 * dij)[pv, :][:, pqpv]
    APV4 = csc_matrix((NPV, NPQPV))

    A = sp_vstack((
        sp_hstack((A1, A2)),
        sp_hstack((APQ3, APQ4)),
        sp_hstack((APV3, APV4))
    )).tocsc()

    print("\ndij_y:\n", dij_y.toarray())
    print("\ndij:\n", dij.toarray())

    print("\nA1:\n", A1.toarray())
    print("\nA2:\n", A2.toarray())

    print("\nAPQ3:\n", APQ3.toarray())
    print("\nAPQ4:\n", APQ4.toarray())

    print("\nAPV3:\n", APV3.toarray())
    print("\nAPV4:\n", APV4.toarray())

    return A, NPQPV


def L2(n, M):
    """
    
    Args:
        n: coefficient order
        M: Reduced set voltages vector for the pv buses

    Returns:
        The L vector
    """
    if n == 0:
        return ones(len(M))
    elif n == 1:
        return 2 * M - 2
    elif n == 2:
        return M * M - 2 * M + 1
    else:
        return zeros(len(M))


def L(n, M):
    """

    Args:
        n: coefficient order
        M: Reduced set voltages vector for the pv buses

    Returns:
        The L vector
    """
    if n == 0:
        return ones(len(M))
    elif n == 1:
        return M - 1
    else:
        return zeros(len(M))


def get_rhs(n, npqpv, V, Y_series, Y_shunt, Sbus, M, pq, pv, pqpv):
    """
    Compute the right hand side vector (rhs)    
    Args:
        n: coefficient order
        V: Voltage coefficients matrix (rows: coeff. order, columns: reduced bus index)
        Y_series: Reduced series admittances matrix
        Y_shunt: Reduced shunt admittances vector
        Sbus: Reduced power injections vector
        M: Reduced set voltages vector for the pv buses
        pq: array of reduced PQ bus numbers
        pv: array of reduced PV bus numbers

    Returns:
        Right hand side vector
    """

    if n == 1:
        r1 = Sbus.real[pqpv] - Y_shunt.real[pqpv]
        rpq = -Sbus.imag[pq] - Y_shunt.imag[pq]
        rpv = L(n, M[pv])**2

    else:

        nbus = Y_series.shape[0]
        val = zeros(nbus, dtype=complex_type)
        valpv = zeros(nbus, dtype=complex_type)
        for m in range(1, n):
            val += conj(V[m, :]) * (Y_series * V[n - m, :])
            valpv += conj(V[m, :]) * V[n - m, :]

        r1 = -val.real[pqpv]
        rpq = -val.imag[pq]
        rpv = -valpv.real[pv] + L(n, M[pv])**2

    return np.hstack((r1, rpq, rpv))


def helmw(Y_series, Y_shunt, Sbus, voltageSetPoints, pq, pv, ref, pqpv, types, eps=1e-3, maxcoefficientCount=50):
    """

    Args:
        Y_series:
        Y_shunt:
        Sbus:
        voltageSetPoints:
        pq:
        pv:
        ref:
        pqpv:
        eps:
        maxcoefficientCount:

    Returns:

    """

    nbus = len(Y_shunt)
    nref = len(ref)

    # reduce the arrays and build the system matrix
    A, npqpv = make_A(Y_series=Y_series, Y_shunt=Y_shunt, pq=pq, pv=pv, pqpv=pqpv, types=types)
    print('\nA:\n', A.toarray())

    # get the set points array
    M = abs(voltageSetPoints)

    # factorize the system matrix only once
    Afac = factorized(A)

    # declare the voltages coefficient matrix
    V = ones((maxcoefficientCount, nbus), dtype=complex_type)

    error = list()
    npv = len(pv)

    for n in range(1, maxcoefficientCount):

        # compute the right hand side of the linear system
        rhs = get_rhs(n, npqpv, V, Y_series, Y_shunt, Sbus, M, pq, pv, pqpv)
        print('\nn:', n, ', rhs:\n', rhs)

        # solve the linear system
        x = Afac(rhs)

        # assign the solution to the voltages (convert floats to complex)
        r, i = np.split(x, 2)
        vn = r + 1j * i

        print('voltage coeff (n):\n', vn)

        # stack the coefficients solution
        V[n, pqpv] = vn

        # compute the voltages with Padè
        # print('\nVoltage coeff: \n', V)
        voltages = voltageSetPoints.copy()
        for j in pqpv:
            voltages[j], _, _ = pade_approximation(n, j, V)

        # print('\nVoltage coeff: \n', V)
        # print('\nVoltage values: \n', voltages)

        # evaluate the solution error F(x0)
        Scalc = voltages * conj(Y_series * voltages)
        mis = Scalc - Sbus  # compute the mismatch
        dx = r_[mis[pv].real, mis[pq].real, mis[pq].imag]  # mismatch in the Jacobian order
        normF = np.linalg.norm(dx, np.Inf)

        if npv > 0:
            a = linalg.norm(mis[pv].real, Inf)
        else:
            a = 0
        b = linalg.norm(mis[pq].real, Inf)
        c = linalg.norm(mis[pq].imag, Inf)
        error.append([a, b, c])

    err_df = pd.DataFrame(array(error), columns=['PV_real', 'PQ_real', 'PQ_imag'])
    err_df.plot(logy=True)

    return voltages, normF


if __name__ == "__main__":
    from GridCal.Engine.calculation_engine import *
    from matplotlib import pyplot as plt
    np.set_printoptions(suppress=True, linewidth=320, formatter={'float': '{: 0.4f}'.format})

    grid = MultiCircuit()
    grid.load_file('lynn5buspq.xlsx')
    # grid.load_file('lynn5buspv.xlsx')
    # grid.load_file('IEEE30.xlsx')
    # grid.load_file('/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE 14.xlsx')
    # grid.load_file('/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE39.xlsx')
    # grid.load_file('/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/1354 Pegase.xlsx')

    grid.compile()

    circuit = grid.circuits[0]

    print('\nYbus:\n', circuit.power_flow_input.Ybus.todense())
    print('\nYseries:\n', circuit.power_flow_input.Yseries.todense())
    print('\nYshunt:\n', circuit.power_flow_input.Yshunt)
    print('\nSbus:\n', circuit.power_flow_input.Sbus)
    print('\nIbus:\n', circuit.power_flow_input.Ibus)
    print('\nVbus:\n', circuit.power_flow_input.Vbus)
    print('\ntypes:\n', circuit.power_flow_input.types)
    print('\npq:\n', circuit.power_flow_input.pq)
    print('\npv:\n', circuit.power_flow_input.pv)
    print('\nvd:\n', circuit.power_flow_input.ref)

    import time
    print('HELM model 4')
    start_time = time.time()
    cmax = 8
    V1, err = helmw(Y_series=circuit.power_flow_input.Yseries,
                    Y_shunt=circuit.power_flow_input.Yshunt,
                    Sbus=circuit.power_flow_input.Sbus,
                    voltageSetPoints=circuit.power_flow_input.Vbus,
                    pq=circuit.power_flow_input.pq,
                    pv=circuit.power_flow_input.pv,
                    ref=circuit.power_flow_input.ref,
                    pqpv=circuit.power_flow_input.pqpv,
                    types=circuit.power_flow_input.types,
                    eps=1e-9,
                    maxcoefficientCount=cmax)

    print("--- %s seconds ---" % (time.time() - start_time))

    print('V module:\t', abs(V1))
    print('V angle: \t', angle(V1))
    print('error: \t', err)

    # check the HELM solution: v against the NR power flow
    print('\nNR')
    options = PowerFlowOptions(SolverType.NR, verbose=False, robust=False, tolerance=1e-9)
    power_flow = PowerFlow(grid, options)

    start_time = time.time()
    power_flow.run()
    print("--- %s seconds ---" % (time.time() - start_time))
    vnr = circuit.power_flow_results.voltage

    print('V module:\t', abs(vnr))
    print('V angle: \t', angle(vnr))
    print('error: \t', circuit.power_flow_results.error)

    # check
    print('\ndiff:\t', V1 - vnr)

    plt.show()



# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.
import time
import multiprocessing
from PySide2.QtCore import QThread, Signal

import numpy as np
import scipy.sparse as sp
from scipy.sparse.linalg import spsolve

from GridCal.Engine.basic_structures import Logger
from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.Simulations.result_types import ResultTypes
from GridCal.Engine.Simulations.results_model import ResultsModel
from GridCal.Engine.Core.snapshot_pf_data import compile_snapshot_circuit


def make_ptdf(Bbus, Bf, pqpv, distribute_slack=True):
    """
    Build the PTDF matrix
    :param Bbus: DC-linear susceptance matrix
    :param Bf: Bus-branch "from" susceptance matrix
    :param pqpv: array of sorted pq and pv node indices
    :param distribute_slack: distribute the slack?
    :return: PTDF matrix. It is a full matrix of dimensions branches x buses
    """

    n = Bbus.shape[0]
    nb = n
    nbi = n
    noref = np.arange(1, nb)
    noslack = pqpv

    if distribute_slack:
        dP = np.ones((n, n)) * (-1 / (n - 1))
        for i in range(n):
            dP[i, i] = 1.0
    else:
        dP = np.eye(n, n)

    # solve for change in voltage angles
    dTheta = np.zeros((nb, nbi))
    Bref = Bbus[noslack, :][:, noref].tocsc()
    dtheta_ref = spsolve(Bref,  dP[noslack, :])

    if sp.issparse(dtheta_ref):
        dTheta[noref, :] = dtheta_ref.toarray()
    else:
        dTheta[noref, :] = dtheta_ref

    # compute corresponding change in branch flows
    # Bf is a sparse matrix
    H = Bf * dTheta

    return H


def make_lodf(Cf, Ct, PTDF, correct_values=True):
    """
    Compute the LODF matrix
    :param Cf: Branch "from" -bus connectivity matrix
    :param Ct: Branch "to" -bus connectivity matrix
    :param PTDF: PTDF matrix in numpy array form (branches, buses)
    :return: LODF matrix of dimensions (branches, branches)
    """
    nl = PTDF.shape[0]

    # compute the connectivity matrix
    Cft = Cf - Ct
    H = PTDF * Cft.T

    # old code
    # h = sp.diags(H.diagonal())
    # LODF = H / (np.ones((nl, nl)) - h * np.ones(nl))

    # divide each row of H by the vector 1 - H.diagonal
    # LODF = H / (1 - H.diagonal())
    #
    # # replace possible nan and inf
    # LODF[LODF == -np.inf] = 0
    # LODF[LODF == np.inf] = 0
    # LODF = np.nan_to_num(LODF)

    # this loop avoids the divisions by zero
    # in those cases the LODF column should be zero
    LODF = np.zeros((nl, nl))
    div = 1 - H.diagonal()
    for j in range(H.shape[1]):
        if div[j] != 0:
            LODF[:, j] = H[:, j] / div[j]

    # replace the diagonal elements by -1
    # old code
    # LODF = LODF - sp.diags(LODF.diagonal()) - sp.eye(nl, nl), replaced by:
    for i in range(nl):
        LODF[i, i] = - 1.0

    if correct_values:
        i1, j1 = np.where(LODF > 1)
        for i, j in zip(i1, j1):
            LODF[i, j] = 1

        i2, j2 = np.where(LODF < -1)
        for i, j in zip(i2, j2):
            LODF[i, j] = -1

    return LODF


class LinearAnalysisResults:

    def __init__(self, n_br=0, n_bus=0, br_names=(), bus_names=(), bus_types=()):
        """
        PTDF and LODF results class
        :param n_br: number of branches
        :param n_bus: number of buses
        :param br_names: branch names
        :param bus_names: bus names
        :param bus_types: bus types array
        """

        self.name = 'Linear Analysis'

        # number of branches
        self.n_br = n_br

        self.n_bus = n_bus

        # names of the branches
        self.br_names = br_names

        self.bus_names = bus_names

        self.bus_types = bus_types

        self.logger = Logger()

        self.PTDF = np.zeros((n_br, n_bus))
        self.LODF = np.zeros((n_br, n_br))

        self.available_results = [ResultTypes.PTDFBranchesSensitivity,
                                  ResultTypes.OTDF]

    def mdl(self, result_type: ResultTypes) -> ResultsModel:
        """
        Plot the results.

        Arguments:

            **result_type**: ResultTypes

        Returns: ResultsModel
        """

        if result_type == ResultTypes.PTDFBranchesSensitivity:
            labels = self.bus_names
            y = self.PTDF
            y_label = '(p.u.)'
            title = 'Branches sensitivity'

        elif result_type == ResultTypes.OTDF:
            labels = self.br_names
            y = self.LODF
            y_label = '(p.u.)'
            title = 'Branch failure sensitivity'

        else:
            labels = []
            y = np.zeros(0)
            y_label = ''
            title = ''

        # assemble model
        mdl = ResultsModel(data=y,
                           index=self.br_names,
                           columns=labels,
                           title=title,
                           ylabel=y_label,
                           units=y_label)
        return mdl


class LinearAnalysis:

    def __init__(self, grid: MultiCircuit, distributed_slack=True, correct_values=True):
        """

        :param grid:
        :param distributed_slack:
        """

        self.grid = grid

        self.distributed_slack = distributed_slack

        self.correct_values = correct_values

        self.numerical_circuit = None

        self.results = LinearAnalysisResults(n_br=0,
                                             n_bus=0,
                                             br_names=[],
                                             bus_names=[],
                                             bus_types=[])

        self.logger = Logger()

    def run(self):
        """
        Run the PTDF and LODF
        """
        self.numerical_circuit = compile_snapshot_circuit(self.grid)
        islands = self.numerical_circuit.split_into_islands()

        self.results = LinearAnalysisResults(n_br=self.numerical_circuit.nbr,
                                             n_bus=self.numerical_circuit.nbus,
                                             br_names=self.numerical_circuit.branch_data.branch_names,
                                             bus_names=self.numerical_circuit.bus_data.bus_names,
                                             bus_types=self.numerical_circuit.bus_data.bus_types)

        # compute the PTDF per islands
        if len(islands) > 0:
            for n_island, island in enumerate(islands):

                # no slacks will make it impossible to compute the PTDF analytically
                if len(island.vd) == 1:
                    if len(island.pqpv) > 0:

                        # compute the PTDF of the island
                        ptdf_island = make_ptdf(Bbus=island.Bbus,
                                                Bf=island.Bf,
                                                pqpv=island.pqpv,
                                                distribute_slack=self.distributed_slack)

                        # assign the PTDF to the matrix
                        self.results.PTDF[np.ix_(island.original_branch_idx, island.original_bus_idx)] = ptdf_island

                        # compute the island LODF
                        lodf_island = make_lodf(Cf=island.Cf,
                                                Ct=island.Ct,
                                                PTDF=ptdf_island,
                                                correct_values=self.correct_values)

                        # assign the LODF to the matrix
                        self.results.LODF[np.ix_(island.original_branch_idx, island.original_branch_idx)] = lodf_island
                    else:
                        self.logger.add_error('No PQ or PV nodes', 'Island {}'.format(n_island))
                else:
                    self.logger.add_error('More than one slack bus', 'Island {}'.format(n_island))
        else:

            # there is only 1 island, compute the PTDF
            self.results.PTDF = make_ptdf(Bbus=islands[0].Bbus,
                                          Bf=islands[0].Bf,
                                          pqpv=islands[0].pqpv,
                                          distribute_slack=self.distributed_slack)

            # compute the LODF upon the PTDF
            self.results.LODF = make_lodf(Cf=islands[0].Cf,
                                          Ct=islands[0].Ct,
                                          PTDF=self.results.PTDF,
                                          correct_values=self.correct_values)

    def get_branch_time_series(self, Sbus):
        """
        Compute the time series PTDF
        :param Sbus: Power injections time series array
        :return:
        """

        # option 2: call the power directly
        P = Sbus.real
        PTDF = self.results.PTDF
        Pbr = np.dot(PTDF, P).T * self.grid.Sbase

        return Pbr



import numpy as np
from numpy import angle, conj, exp, r_, Inf
from numpy.linalg import norm
from scipy.sparse.linalg import splu
import time
from GridCal.Engine.Simulations.PowerFlow.power_flow_results import NumericPowerFlowResults

np.set_printoptions(linewidth=320)


def FDPF(Vbus, Sbus, Ibus, Ybus, B1, B2, pq, pv, pqpv, tol=1e-9, max_it=100) -> NumericPowerFlowResults:
    """
    Fast decoupled power flow
    :param Vbus:
    :param Sbus:
    :param Ibus:
    :param Ybus:
    :param B1:
    :param B2:
    :param pq:
    :param pv:
    :param pqpv:
    :param tol:
    :param max_it:
    :return:
    """

    start = time.time()
    # pvpq = np.r_[pv, pq]

    # set voltage vector for the iterations
    voltage = Vbus.copy()
    Va = np.angle(voltage)
    Vm = np.abs(voltage)

    # Factorize B1 and B2
    J1 = splu(B1[np.ix_(pqpv, pqpv)])
    J2 = splu(B2[np.ix_(pq, pq)])

    # evaluate initial mismatch
    Scalc = voltage * np.conj(Ybus * voltage - Ibus)
    mis = (Scalc - Sbus) / Vm  # complex power mismatch
    dP = mis[pqpv].real
    dQ = mis[pq].imag

    if len(pqpv) > 0:
        normP = norm(dP, Inf)
        normQ = norm(dQ, Inf)
        converged = normP < tol and normQ < tol

        # iterate
        iter_ = 0
        while not converged and iter_ < max_it:

            iter_ += 1

            # ----------------------------- P iteration to update Va ----------------------
            # solve voltage angles
            dVa = J1.solve(dP)

            # update voltage
            Va[pqpv] -= dVa
            voltage = Vm * exp(1j * Va)

            # evaluate mismatch
            Scalc = voltage * conj(Ybus * voltage - Ibus)
            mis = (Scalc - Sbus) / Vm  # complex power mismatch
            dP = mis[pqpv].real
            dQ = mis[pq].imag
            normP = norm(dP, Inf)
            normQ = norm(dQ, Inf)

            if normP < tol and normQ < tol:
                converged = True
            else:
                # ----------------------------- Q iteration to update Vm ----------------------
                # Solve voltage modules
                dVm = J2.solve(dQ)

                # update voltage
                Vm[pq] -= dVm
                voltage = Vm * exp(1j * Va)

                # evaluate mismatch
                Scalc = voltage * conj(Ybus * voltage - Ibus)
                mis = (Scalc - Sbus) / Vm  # complex power mismatch
                dP = mis[pqpv].real
                dQ = mis[pq].imag
                normP = norm(dP, Inf)
                normQ = norm(dQ, Inf)

                if normP < tol and normQ < tol:
                    converged = True

        F = r_[dP, dQ]  # concatenate again
        normF = norm(F, Inf)

    else:
        converged = True
        iter_ = 0
        normF = 0

    end = time.time()
    elapsed = end - start

    return NumericPowerFlowResults(voltage, converged, normF, Scalc, None, None, None, iter_, elapsed)


if __name__ == '__main__':

    fname = r'/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE 9 Bus.gridcal'

    from GridCal.Engine import FileOpen

    circuit = FileOpen(fname).open()
    nc = circuit.compile_snapshot()
    islands = nc.compute()
    island = islands[0]

    voltage_, converged_, normF_, Scalc_, iter_, elapsed_ = FDPF(Vbus=island.Vbus,
                                                                 Sbus=island.Sbus,
                                                                 Ibus=island.Ibus,
                                                                 Ybus=island.Ybus,
                                                                 B1=island.B1,
                                                                 B2=island.B2,
                                                                 pq=island.pq,
                                                                 pv=island.pv,
                                                                 pqpv=island.pqpv,
                                                                 tol=1e-9,
                                                                 max_it=100)

    print(np.abs(voltage_))
    print('iter:', iter_)
    print('Error:', normF_)



''' Calculate e using its simple continued fraction expansion

    See http://stackoverflow.com/q/36077810/4014959

    Also see
    https://en.wikipedia.org/wiki/Continued_fraction#Regular_patterns_in_continued_fractions

    Written by PM 2Ring 2016.03.18
'''

from __future__ import print_function, division
import sys

def contfrac_to_frac(seq):
    ''' Convert the simple continued fraction in `seq`
        into a fraction, num / den
    '''
    num, den = 1, 0
    for u in reversed(seq):
        num, den = den + num*u, num
    return num, den

def e_cont_frac(n):
    ''' Build `n` terms of the simple continued fraction expansion of e
        `n` must be a positive integer
    '''
    seq = [2 * (i+1) // 3 if i%3 == 2 else 1 for i in range(n)]
    seq[0] += 1
    return seq


def main():

    # Get the the number of terms, less one
    n = int(sys.argv[1]) if len(sys.argv) > 1 else 11
    if n < 0:
        print('Argument must be >= 0')
        exit()

    n += 1
    seq = e_cont_frac(n)
    num, den = contfrac_to_frac(seq)

    print('Terms =', n)
    print('Continued fraction:', seq)
    print('Fraction: {0} / {1}'.format(num, den))
    print('Float {0:0.15f}'.format(num / den))

if __name__ == '__main__':
    main()


import copy

"""
Pure Python/Numpy implementation of the Nelder-Mead algorithm from
https://github.com/alxsoares/nelder-mead/blob/master/nelder_mead.py
Reference: https://en.wikipedia.org/wiki/Nelder%E2%80%93Mead_method

transformations from https://github.com/alexblaessle/constrNMPy/blob/master/constrNMPy/constrNMPy.py
"""


def transform_x(x, LB, UB, offset=1E-20):
    """Transforms ``x`` into constrained form, obeying upper bounds ``UB`` and lower bounds ``LB``.
    .. note:: Will add tiny offset to LB if ``LB[i]=0``, to avoid singularities.
    Idea taken from http://www.mathworks.com/matlabcentral/fileexchange/8277-fminsearchbnd--fminsearchcon
    Args:
        x (numpy.ndarray): Input vector.
        LB (numpy.ndarray): Lower bounds.
        UB (numpy.ndarray): Upper bounds.
    Keyword Args:
        offset (float): Small offset added to lower bound if LB=0.
    Returns:
        numpy.ndarray: Transformed x-values.
    """

    # Make sure everything is float
    x = np.asarray(x, dtype=np.float64)

    # Add offset if necessary to avoid singularities
    for l in LB:
        if l == 0:
            l = l + offset

    # Determine number of parameters to be fitted
    nparams = len(x)

    # Make empty vector
    xtrans = np.zeros(np.shape(x))

    # k allows some variables to be fixed, thus dropped from the
    # optimization.
    k = 0

    for i in range(nparams):

        # Upper bound only
        if UB[i] is not None and LB[i] is None:

            xtrans[i] = UB[i] - x[k] ** 2
            k = k + 1

        # Lower bound only
        elif UB[i] is None and LB[i] is not None:

            xtrans[i] = LB[i] + x[k] ** 2
            k = k + 1

        # Both bounds
        elif UB[i] is not None and LB[i] is not None:

            xtrans[i] = (np.sin(x[k]) + 1.) / 2. * (UB[i] - LB[i]) + LB[i]
            xtrans[i] = max([LB[i], min([UB[i], xtrans[i]])])
            k = k + 1

        # No bounds
        elif UB[i] is None and LB[i] is None:

            xtrans[i] = x[k]
            k = k + 1

        # NOTE: The original file has here another case for fixed variable. We might need to add this here!!!

    return np.array(xtrans)


def transform_x0(x0, LB, UB):
    r"""Transforms ``x0`` into constrained form, obeying upper bounds ``UB`` and lower bounds ``LB``.
    Idea taken from http://www.mathworks.com/matlabcentral/fileexchange/8277-fminsearchbnd--fminsearchcon
    Args:
        x0 (numpy.ndarray): Input vector.
        LB (numpy.ndarray): Lower bounds.
        UB (numpy.ndarray): Upper bounds.
    Returns:
        numpy.ndarray: Transformed x-values.
    """

    # Turn into list
    x0u = list(x0)

    k = 0
    for i in range(len(x0)):

        # Upper bound only
        if UB[i] is not None and LB[i] is None:
            if UB[i] <= x0[i]:
                x0u[k] = 0
            else:
                x0u[k] = np.sqrt(UB[i] - x0[i])
            k = k + 1

        # Lower bound only
        elif UB[i] is None and LB[i] is not None:
            if LB[i] >= x0[i]:
                x0u[k] = 0
            else:
                x0u[k] = np.sqrt(x0[i] - LB[i])
            k = k + 1

        # Both bounds
        elif UB[i] is not None and LB[i] is not None:
            if UB[i] <= x0[i]:
                x0u[k] = np.pi / 2
            elif LB[i] >= x0[i]:
                x0u[k] = -np.pi / 2
            else:
                x0u[k] = 2 * (x0[i] - LB[i]) / (UB[i] - LB[i]) - 1;
                # shift by 2*pi to avoid problems at zero in fmin otherwise, the initial simplex is vanishingly small
                x0u[k] = 2 * np.pi + np.arcsin(max([-1, min(1, x0u[k])]));
            k = k + 1

        # No bounds
        elif UB[i] is None and LB[i] is None:
            x0u[k] = x0[i]
            k = k + 1

    return np.array(x0u)


def nelder_mead(objective_function, x_start, LB, UB,
                step=0.1, no_improve_thr=10e-6,
                no_improv_break=10, max_iter=0,
                alpha=1., gamma=2., rho=-0.5, sigma=0.5):
    """
        @param f (function): function to optimize, must return a scalar score
            and operate over a numpy array of the same dimensions as x_start
        @param x_start (numpy array): initial position
        @param step (float): look-around radius in initial step
        @no_improv_thr,  no_improv_break (float, int): break after no_improv_break iterations with
            an improvement lower than no_improv_thr
        @max_iter (int): always break after this number of iterations.
            Set it to 0 to loop indefinitely.
        @alpha, gamma, rho, sigma (floats): parameters of the algorithm
            (see Wikipedia page for reference)
        return: tuple (best parameter array, best score)
    """

    # init
    dim = len(x_start)
    prev_best = objective_function(transform_x(x_start, LB, UB))
    no_improv = 0
    res = [[x_start, prev_best]]

    for i in range(dim):
        x = copy.copy(x_start)
        x[i] = x[i] + step
        score = objective_function(transform_x(x, LB, UB))
        res.append([x, score])

    # simplex iter
    iters = 0
    while 1:
        # order
        res.sort(key=lambda x: x[1])
        best = res[0][1]

        # break after max_iter
        if max_iter and iters >= max_iter:
            return res[0]
        iters += 1

        # break after no_improv_break iterations with no improvement
        print('...best so far:', best)

        if best < prev_best - no_improve_thr:
            no_improv = 0
            prev_best = best
        else:
            no_improv += 1

        if no_improv >= no_improv_break:
            return res[0]

        # centroid
        x0 = [0.] * dim
        for tup in res[:-1]:
            for i, c in enumerate(tup[0]):
                x0[i] += c / (len(res)-1)

        # reflection
        xr = x0 + alpha*(x0 - res[-1][0])
        rscore = objective_function(transform_x(xr, LB, UB))
        if res[0][1] <= rscore < res[-2][1]:
            del res[-1]
            res.append([xr, rscore])
            continue

        # expansion
        if rscore < res[0][1]:
            xe = x0 + gamma*(x0 - res[-1][0])
            escore = objective_function(transform_x(xe, LB, UB))
            if escore < rscore:
                del res[-1]
                res.append([xe, escore])
                continue
            else:
                del res[-1]
                res.append([xr, rscore])
                continue

        # contraction
        xc = x0 + rho*(x0 - res[-1][0])
        cscore = objective_function(transform_x(xc, LB, UB))
        if cscore < res[-1][1]:
            del res[-1]
            res.append([xc, cscore])
            continue

        # reduction
        x1 = res[0][0]
        nres = []
        for tup in res:
            redx = x1 + sigma*(tup[0] - x1)
            score = objective_function(transform_x(redx, LB, UB))
            nres.append([redx, score])
        res = nres


if __name__ == "__main__":
    # test
    import math
    import numpy as np

    def f(x):
        return math.sin(x[0]) * math.cos(x[1]) * (1. / (abs(x[2]) + 1))

    x0 = np.array([0., 0., 0.])
    LB = np.ones_like(x0) * -1
    UB = np.ones_like(x0) * 1
    result = nelder_mead(f, x0, LB, UB)

    print('Result')
    print(result)